# -*- coding: utf-8 -*-
'''
.. module:: skrf.circuit
========================================
circuit (:mod:`skrf.circuit`)
========================================

Circuit class represents a circuit of arbitrary topology,
consisting of an arbitrary number of N-ports networks. 

Like in an electronic circuit simulator, the circuit must have one or more ports 
connected to the circuit. The Circuit object allows one retrieving the M-ports network, 
where M is the number of ports defined. 

The results are returned in :class:`~skrf.network.Circuit` object.

   
Building a Circuit 
==================
.. autosummary::
   :toctree: generated/

   Circuit
   Circuit.Port
   Circuit.Ground

Representing a Circuit
=======================
.. autosummary::
   :toctree: generated/
   
   Circuit.plot_graph

Network Representations
========================
.. autosummary::
   :toctree: generated/
   
   Circuit.network
   Circuit.s
   Circuit.s_external
   Circuit.s_active
   Circuit.z_active
   Circuit.y_active
   Circuit.vswr_active
   Circuit.port_z0

Circuit internals
==================
.. autosummary::
   :toctree: generated/
   
   Circuit.networks_dict
   Circuit.networks_list
   Circuit.connections_nb
   Circuit.connections_list
   Circuit.nodes_nb
   Circuit.dim
   Circuit.intersections_dict
   Circuit.port_indexes
   Circuit.C
   Circuit.X

Graph representation
========================
.. autosummary::
   :toctree: generated/
   
   Circuit.graph
   Circuit.G
   Circuit.edges
   Circuit.edge_labels

'''
from . network import Network, a2s
from . media import media
from . constants import INF

import numpy as np

try:
    import networkx as nx
except ImportError as e:
    pass

from itertools import chain, product
from scipy.linalg import block_diag

class Circuit():
    '''
    Creates a circuit made of a set of N-ports networks.
    
    For instructions on how to create Circuit see  :func:`__init__`.
    
    A Circuit object is representation a circuit assembly of an arbitrary 
    number of N-ports networks connected together via an arbitrary topology.

    The algorithm used to calculate the resultant network can be found in [#]_.

    References
    ----------
    .. [#] P. Hallbjörner, Microw. Opt. Technol. Lett. 38, 99 (2003).

    '''
    def __init__(self, connections):
        '''
        Circuit constructor. Creates a circuit made of a set of N-ports networks.

        Parameters
        -----------
        connections : list of list of tuples
            Description of circuit connections.
            Each connection is a described by a list of tuple. 
            Each tuple contains (network, network_port_nb). 
            Port number indexing starts from zero.
            
        
        Example
        ------------
        Example of connections between two 1-port networks:
        ::
            connections = [ 
                [(network1, 0), (network2, 0)],
            ]
        
        Example of a connection between three 1-port networks connected 
        to a single node:
        ::
            connections = [ 
                [(network1, 0), (network2, 0), (network3, 0)]
            ]
            
        Example of a connection between two 1-port networks (port1 and port2)
        and two 2-ports networks (ntw1 and ntw2):
        ::
            connections = [
                [(port1, 0), (ntw1, 0)],
                [(ntw1, 1), (ntw2, 0)],
                [(ntw2, 1), (port2, 0)]
            ]

        Example of a connection between three 1-port networks (port1, port2 and port3)
        and a 3-ports network (ntw):
        ::
            connections = [
                [(port1, 0), (ntw, 0)],
                [(port2, 0), (ntw, 1)],
                [(port3, 0), (ntw, 2)]
            ]

        NB1: Creating 1-port network to be used a port can be made with :func:`Port`

        NB2: The external ports indexing is defined by the order of appearance of
        the ports in the connections list. Thus, the first network identified 
        as a port will be the 1st port of the resulting network (index 0), 
        the second network identified as a port will be the second port (index 1), 
        etc.


        '''
        self.connections = connections

        # check if all networks have a name
        for cnx in self.connections:
            for (ntw, _) in cnx:
                if not self._is_named(ntw):
                    raise AttributeError('All Networks must have a name. Faulty network:', ntw)

        # list of networks for initial checks
        ntws = self.networks_list()

        # check if all networks have same frequency
        ref_freq = ntws[0].frequency
        for ntw in ntws:
            if ntw.frequency != ref_freq:
                raise AttributeError('All Networks must have same frequencies')
        # All frequencies are the same, Circuit frequency can be any of the ntw
        self.frequency = ntws[0].frequency

    def _is_named(self, ntw):
        '''
        Return True is the network has a name, False otherwise
        '''
        if not ntw.name or ntw.name == '':
            return False
        else:
            return True

    @classmethod
    def Port(cls, frequency, name, z0=50):
        '''
        Return a 1-port Network to be used as a Circuit port. 
        
        Passing the frequency and name is mendatory. Port name must include 
        the word 'port' inside. (ex: 'Port1' or 'port_3')

        Parameters
        -----------
        frequency : :class:`~skrf.frequency.Frequency`
            Frequency common to all other networks in the circuit
        name : string
            Name of the port.
            Must inlude the word 'port' inside. (ex: 'Port1' or 'port_3')
        z0 : real
            Characteristic impedance of the port. Default is 50 Ohm.

        Return
        --------
        port : :class:`~skrf.network.Network` object
            (External) 1-port network

        Examples
        -----------
        .. ipython::

            @suppress
            In [16]: import skrf as rf

            In [17]: freq = rf.Frequency(start=1, stop=2, npoints=101)

            In [18]: port1 = rf.Circuit.Port(freq, name='Port1')
        '''
        _media = media.DefinedGammaZ0(frequency, z0=z0)
        return _media.match(name=name)

    @classmethod
    def Ground(cls, frequency, name, z0=50):
        '''
        Return a 1-port network of a grounded link, to be used in a Circuit description. 
        
        Passing the frequency and name is mendatory.
        
        The ground link is modelled as an infinite admittance.
        
        Parameters
        -----------
        frequency : :class:`~skrf.frequency.Frequency`
            Frequency common to all other networks in the circuit
        name : string
            Name of the ground.
        z0 : real
            Characteristic impedance of the port. Default is 50 Ohm.

        Return
        --------
        ground : :class:`~skrf.network.Network` object
            (External) 2-port network

        Examples
        -----------
        .. ipython::

            @suppress
            In [16]: import skrf as rf

            In [17]: freq = rf.Frequency(start=1, stop=2, npoints=101)

            In [18]: ground = rf.Circuit.Ground(freq, name='GND')
        
        '''
        Y = INF
        A = np.zeros(shape=(len(frequency), 2, 2), dtype=complex)
        A[:, 0, 0] = 1
        A[:, 0, 1] = 0
        A[:, 1, 0] = Y
        A[:, 1, 1] = 1
        ntw = Network(frequency=frequency, z0=z0, name=name)
        ntw.s = a2s(A)
        return ntw

    def networks_dict(self, connections=None, min_nports=1):
        '''
        Return the dictionnary of Networks from the connection setup X
        '''
        if not connections:
            connections = self.connections

        ntws = []
        for cnx in connections:
            for (ntw, port) in cnx:
                ntws.append(ntw)
        return {ntw.name: ntw for ntw in ntws  if ntw.nports >= min_nports}

    def networks_list(self, connections=None, min_nports=1):
        '''
        Return a list of unique networks (sorted by appearing order in connections)
        '''
        if not connections:
            connections = self.connections

        ntw_dict = self.networks_dict(connections)
        return [ntw for ntw in ntw_dict.values() if ntw.nports >= min_nports]

    @property
    def connections_nb(self):
        '''
        Return the number of intersections in the circuit.
        '''
        return len(self.connections)

    @property
    def connections_list(self):
        '''
        Return the full list of connections, including intersections.
        The resulting list if of the form:
        ::
            [
             [connexion_number, connexion],
             [connexion_number, connexion],
             ...
            ]
        '''
        return [[idx_cnx, cnx] for (idx_cnx, cnx) in enumerate(chain.from_iterable(self.connections))]



    @property
    def networks_nb(self):
        '''
        Return the number of connected networks (port excluded)
        '''
        return len(self.networks_list(self.connections))

    @property
    def nodes_nb(self):
        '''
        Return the number of nodes in the circuit.
        '''
        return self.connections_nb + self.networks_nb

    @property
    def dim(self):
        '''
        Return the dimension of the C, X and global S matrices.

        It correspond to the sum of all connections
        '''
        return np.sum([len(cnx) for cnx in self.connections])

    @property
    def G(self):
        '''
        Generate the graph of the circuit. Same as :func:`graph`.
        '''
        return self.graph()

    def graph(self):
        '''
        Generate the graph of the circuit.
        
        Return
        ------
        G: `networkx  <https://networkx.github.io/>`_ 
            graph object.
        
        
        '''
        try:
            import networkx as nx
        except ImportError as e:
            raise ImportError('networkx package as not been installed and is required. ')

        G = nx.Graph()
        # Adding network nodes
        G.add_nodes_from([it for it in self.networks_dict(self.connections)])

        # Adding edges in the graph between connections and networks
        for (idx, cnx) in enumerate(self.connections):
            cnx_name = 'X'+str(idx)
            # Adding connection nodes and edges
            G.add_node(cnx_name)
            for (ntw, ntw_port) in cnx:
                ntw_name = ntw.name
                G.add_edge(cnx_name, ntw_name)
        return G

    def is_connected(self):
        '''
        Check if the circuit's graph is connected, that is if
        if every pair of vertices in the graph is connected.
        '''
        # Get the circuit graph. Will raise an error if the networkx package
        # is not installed. 
        G = self.G
    
        return nx.algorithms.components.is_connected(G)

    @property
    def intersections_dict(self):
        '''
        Return a dictionnary of all intersections with associated ports and z0:
        ::
            { k: [(ntw1_name, ntw1_port), (ntw1_z0, ntw2_name, ntw2_port), ntw2_z0], ... }
        '''
        inter_dict = {}
        # for k in range(self.connections_nb):
        #     # get all edges connected to intersection Xk
        #     inter_dict[k] = list(nx.algorithms.boundary.edge_boundary(self.G, ('X'+str(k),) ))

        for (k, cnx) in enumerate(self.connections):
            inter_dict[k] = [(ntw, ntw_port, ntw.z0[0, ntw_port]) \
                                  for (ntw, ntw_port) in cnx]
        return inter_dict

    @property
    def edges(self):
        '''
        Return the list of all circuit connections
        '''
        return list(self.G.edges)

    @property
    def edge_labels(self):
        '''
        Return a dictionnary describing the port and z0 of all graph edges.

        Dictionnary is in the form:
        ::
            {('ntw1_name', 'X0'): '3 (50+0j)',
             ('ntw2_name', 'X0'): '0 (50+0j)',
             ('ntw2_name', 'X1'): '2 (50+0j)', ... }
        
        which can be used in `networkx.draw_networkx_edge_labels`
        '''
        # for all intersections, 
        # get the N interconnected networks and associated ports and z0
        # and forge the edge label dictionnary containing labels between
        # two nodes
        edge_labels = {}
        for it in self.intersections_dict.items():
            k, cnx = it
            for idx in range(len(cnx)):
                ntw, ntw_port, ntw_z0 = cnx[idx]
                #ntw_z0 = ntw.z0[0,ntw_port]
                edge_labels[(ntw.name, 'X'+str(k))] = str(ntw_port)+'\n'+\
                                        str(np.round(ntw_z0, decimals=1))

        return edge_labels


    def _Y_k(self, cnx):
        '''
        Return the sum of the system admittances of each intersection
        '''
        y_ns = []
        for (ntw, ntw_port) in cnx:
            # formula (2)
            y_ns.append(1/ntw.z0[:,ntw_port] )

        y_k = np.array(y_ns).sum(axis=0)  # shape: (nb_freq,)
        return y_k

    def _Xnn_k(self, cnx_k):
        '''
        Return the reflection coefficients x_nn of the connection matrix [X]_k
        '''
        X_nn = []
        y_k = self._Y_k(cnx_k)

        for (ntw, ntw_port) in cnx_k:
            # formula (1)
            X_nn.append( 2/(ntw.z0[:,ntw_port]*y_k) - 1)

        return np.array(X_nn).T  # shape: (nb_freq, nb_n)

    def _Xmn_k(self, cnx_k):
        '''
        Return the transmission coefficient x_mn of the mth column of
        intersection scattering matrix matrix [X]_k
        '''
        # get the char.impedance of the n
        X_mn = []
        y_k = self._Y_k(cnx_k)

        # There is a problem in the case of two-ports connexion:
        # the formula (3) in P. Hallbjörner (2003) seems incorrect.
        # Instead of Z_n one should have sqrt(Z_1 x Z_2).
        # The formula works with respect to the example given in the paper
        # (3 port connection), but not with 2-port connections made with skrf
        if len(cnx_k) == 2:
            z0s = []
            for (ntw, ntw_port) in cnx_k:
                z0s.append(ntw.z0[:,ntw_port])

            z0eq = np.array(z0s).prod(axis=0)

            for (ntw, ntw_port) in cnx_k:
                X_mn.append( 2/(np.sqrt(z0eq) *y_k) )
        else:
            # formula (3)
            for (ntw, ntw_port) in cnx_k:
                X_mn.append( 2/(ntw.z0[:,ntw_port]*y_k) )

        return np.array(X_mn).T  # shape: (nb_freq, nb_n)

    def _Xk(self, cnx_k):
        '''
        Return the scattering matrices [X]_k of the individual intersections k
        '''
        Xnn = self._Xnn_k(cnx_k)  # shape: (nb_freq, nb_n)
        Xmn = self._Xmn_k(cnx_k)  # shape: (nb_freq, nb_n)
        # # for loop version
        # Xs = []
        # for (_Xnn, _Xmn) in zip(Xnn, Xmn):  # for all frequencies
        #       # repeat Xmn along the lines
        #     _X = np.tile(_Xmn, (len(_Xmn), 1))
        #     _X[np.diag_indices(len(_Xmn))] = _Xnn
        #     Xs.append(_X)

        # return np.array(Xs) # shape : nb_freq, nb_n, nb_n

        # vectorized version
        nb_n = Xnn.shape[1]
        Xs = np.tile(Xmn, (nb_n, 1, 1)).swapaxes(1, 0)
        Xs[:, np.arange(nb_n), np.arange(nb_n)] = Xnn

        return Xs # shape : nb_freq, nb_n, nb_n

        # TEST : Could we use media.splitter() instead ? -> does not work
        # _media = media.DefinedGammaZ0(frequency=self.frequency)
        # Xs = _media.splitter(len(cnx_k), z0=self._cnx_z0(cnx_k))
        # return Xs.s

    @property
    def X(self):
        '''
        Return the concatenated intersection matrix [X] of the circuit.

        It is composed of the individual intersection matrices [X]_k assembled
        by bloc diagonal.
        '''
        # Xk = []
        # for cnx in self.connections:
        #     Xk.append(self._Xk(cnx))
        # Xk = np.array(Xk)

        # #X = np.zeros(len(C.frequency), )
        # Xf = []
        # for (idx, f) in enumerate(self.frequency):
        #     Xf.append(block_diag(*Xk[:,idx,:]))
        # return np.array(Xf)  # shape: (nb_frequency, nb_inter*nb_n, nb_inter*nb_n)

        # Slightly faster version
        Xks = [self._Xk(cnx) for cnx in self.connections]

        Xf = np.zeros((len(self.frequency), self.dim, self.dim), dtype='complex')
        # TODO: avoid this for loop which is a bottleneck for large frequencies
        for idx in np.nditer(np.arange(len(self.frequency))):
            mat_list = [Xk[idx,:] for Xk in Xks]
            Xf[idx,:] = block_diag(*mat_list)  # bottleneck 
        return Xf

    @property
    def C(self):
        '''
        Return the global scattering matrix of the networks
        '''
        # list all networks which are not considered as "ports", 
        # that is which do not contain "port" in their network name
        ntws = {k:v for k,v in self.networks_dict().items() if 'port' not in k.lower()}

        # generate the port reordering indexes from each connections
        ntws_ports_reordering = {ntw:[] for ntw in ntws}
        for (idx_cnx, cnx) in self.connections_list:
            ntw, ntw_port = cnx
            if ntw.name in ntws.keys():
                ntws_ports_reordering[ntw.name].append([ntw_port, idx_cnx])

        # re-ordering scattering parameters
        S = np.zeros((len(self.frequency), self.dim, self.dim), dtype='complex' )

        for (ntw_name, ntw_ports) in ntws_ports_reordering.items():
            # get the port re-ordering indexes (from -> to)
            ntw_ports = np.array(ntw_ports)
            # create the port permutations
            from_port = list(product(ntw_ports[:,0], repeat=2))
            to_port = list(product(ntw_ports[:,1], repeat=2))

            #print(ntw_name, from_port, to_port)
            for (_from, _to) in zip(from_port, to_port):
                #print(f'{_from} --> {_to}')
                S[:, _to[0], _to[1]] = ntws[ntw_name].s[:, _from[0], _from[1]]

        return S  # shape (nb_frequency, nb_inter*nb_n, nb_inter*nb_n)

    @property
    def s(self):
        '''
        Return the global scattering parameters of the circuit, that is with
        both "inner" and "outer" ports
        '''
        # transpose is necessary to get expected result
        #return np.transpose(self.X @ np.linalg.inv(np.identity(self.dim) - self.C @ self.X), axes=(0,2,1))
        # does not use the @ operator for backward Python version compatibility
        return np.transpose(np.matmul(self.X, np.linalg.inv(np.identity(self.dim) - np.matmul(self.C, self.X))), axes=(0,2,1))


    @property
    def port_indexes(self):
        '''
        Return the indexes of the "external" ports. These must be labelled "port"
        '''
        port_indexes = []
        for (idx_cnx, cnx) in enumerate(chain.from_iterable(self.connections)):
            ntw, ntw_port = cnx
            if 'port' in str.lower(ntw.name):
                port_indexes.append(idx_cnx)
        return port_indexes

    def _cnx_z0(self, cnx_k):
        '''
        Return the characteristic impedances of a specific connection
        '''
        z0s = []
        for (ntw, ntw_port) in cnx_k:
            z0s.append(ntw.z0[:,ntw_port])

        return np.array(z0s).T  # shape (nb_freq, nb_ports_at_cnx)

    @property
    def port_z0(self):
        '''
        Return the external port impedances
        '''
        z0s = []
        for cnx in self.connections:
            for (ntw, ntw_port) in cnx:
                z0s.append(ntw.z0[:,ntw_port])

        return np.array(z0s)[self.port_indexes, :].T  # shape (nb_freq, nb_ports)

    @property
    def s_external(self):
        '''
        Return the scattering parameters for the external ports
        '''
        port_indexes = self.port_indexes
        a, b = np.meshgrid(port_indexes, port_indexes)
        S_ext = self.s[:, a, b]
        return S_ext  # shape (nb_frequency, nb_ports, nb_ports)

    @property
    def network(self):
        '''
        Return the Network associated to external ports
        '''
        ntw = Network()
        ntw.frequency = self.frequency
        ntw.z0 = self.port_z0
        ntw.s = self.s_external
        return ntw


    def s_active(self, a):
        '''
        Return active s-parameters of the circuit's network for a defined wave excitation a.
        
        The active s-parameter at a port is the reflection coefficients 
        when other ports are excited. It is an important quantity for active
        phased array antennas.
        
        Active s-parameters are defined by [#]_:
        
        .. math::
                    
            \mathrm{active}(s)_{mn} = \sum_i s_{mi} \\frac{a_i}{a_n}
        
        Parameters
        ----------
        a : complex array of shape (n_ports)
            forward wave complex amplitude (pseudowave formulation [#]_)
        
        Return
        ---------
        s_act : complex array of shape (n_freqs, n_ports)
            active s-parameters for the excitation a
        
        
        References
        ---------- 
        .. [#] D. M. Pozar, IEEE Trans. Antennas Propag. 42, 1176 (1994).
        
        .. [#] D. Williams, IEEE Microw. Mag. 14, 38 (2013).
        
        '''
        return self.network.s_active(a)

    def z_active(self, a):
        '''
        Return the active Z-parameters of the circuit's network for a defined wave excitation a.
        
        The active Z-parameters are defined by:
            
        .. math::
                    
            \mathrm{active}(z)_{m} = z_{0,m} \\frac{1 + \mathrm{active}(s)_m}{1 - \mathrm{active}(s)_m}
            
        where :math:`z_{0,m}` is the characteristic impedance and
        :math:`\mathrm{active}(s)_m` the active S-parameter of port :math:`m`.
        
        Parameters
        ----------
        a : complex array of shape (n_ports)
            forward wave complex amplitude
    
        Return
        ----------
        z_act : complex array of shape (nfreqs, nports)
            active Z-parameters for the excitation a
            
        See Also
        -----------
            s_active : active S-parameters
            y_active : active Y-parameters
            vswr_active : active VSWR        
        '''
        return self.network.z_active(a)

    def y_active(self, a):
        '''
        Return the active Y-parameters of the circuit's network for a defined wave excitation a.
        
        The active Y-parameters are defined by:
            
        .. math::
                    
            \mathrm{active}(y)_{m} = y_{0,m} \\frac{1 - \mathrm{active}(s)_m}{1 + \mathrm{active}(s)_m}
            
        where :math:`y_{0,m}` is the characteristic admittance and
        :math:`\mathrm{active}(s)_m` the active S-parameter of port :math:`m`.    
        
        Parameters
        ----------            
        a : complex array of shape (n_ports)
            forward wave complex amplitude 
        
        Return
        ----------
        y_act : complex array of shape (nfreqs, nports)
            active Y-parameters for the excitation a
            
        See Also
        -----------
            s_active : active S-parameters
            z_active : active Z-parameters
            vswr_active : active VSWR       
        '''
        return self.network.y_active(a)

    def vswr_active(self, a):
        '''
        Return the active VSWR of the circuit's network for a defined wave excitation a.
        
        The active VSWR is defined by :
            
        .. math::
                    
            \mathrm{active}(vswr)_{m} = \\frac{1 + |\mathrm{active}(s)_m|}{1 - |\mathrm{active}(s)_m|}
    
        where :math:`\mathrm{active}(s)_m` the active S-parameter of port :math:`m`.
        
        Parameters
        ----------       
        a : complex array of shape (n_ports)
            forward wave complex amplitude
    
        Return
        ----------
        vswr_act : complex array of shape (nfreqs, nports)
            active VSWR for the excitation a
            
        See Also
        -----------
            s_active : active S-parameters
            z_active : active Z-parameters
            y_active : active Y-parameters  
        '''        
        return self.network.vswr_active(a)
    
    @property
    def z0(self):
        '''
        Characteristic impedances of "internal" ports

        Returns
        -------
        z0 : complex array of shape (nfreqs, nports)
            Characteristic impedances of both "inner" and "outer" ports

        '''
        z0s = []
        for cnx_idx, (ntw, ntw_port) in self.connections_list:
            z0s.append(ntw.z0[:,ntw_port])
        return np.array(z0s).T

    @property
    def connections_pair(self):
        '''
        List the connections by pair. 
        
        Each connection in the circuit is between a specific pair of two 
        (networks, port, z0).  

        Returns
        -------
        connections_pair : list
            list of pair of connections

        '''
        return [self.connections_list[i:i+2] for i in range(0, len(self.connections_list), 2)]
        
    
    @property
    def _currents_directions(self):
        '''
        Create a array of indices to define the sign of the current.
        
        The currents are defined positive when entering an internal network.
        
        

        Returns
        -------
        directions : array of int (nports, 2)
        
        Note
        ----
        This function is used in internal currents and voltages calculations.

        '''
        directions = np.zeros((self.dim,2), dtype='int')
        for cnx_pair in self.connections_pair:
            (cnx_idx_A, cnx_A), (cnx_idx_B, cnx_B) = cnx_pair
            directions[cnx_idx_A,:] = cnx_idx_A, cnx_idx_B
            directions[cnx_idx_B,:] = cnx_idx_B, cnx_idx_A
        return directions
    
    def _a(self, a_external):
        '''
        Wave input array at "internal" ports

        Parameters
        ----------
        a_external : array
            power-wave input vector at ports
            
        Returns
        -------
        a_internal : array
            Wave input array at internal ports

        '''
        # create a zero array and fill the values corresponding to ports
        a_internal = np.zeros(self.dim, dtype='complex')
        a_internal[self.port_indexes] = a_external
        return a_internal

    def _a_external(self, power, phase):
        '''
        Wave input array at Circuit's ports ("external" ports).
        
        The array is defined from power and phase by:
        
        .. math::
            
            a = \sqrt(2 P_{in} ) e^{j \phi}
            
        The factor 2 is in order to deal with peak values.

        Parameters
        ----------
        power : list or array
            Input power at external ports in Watts [W]
        phase : list or array
            Input phase at external ports in radian [rad]

        NB: the size of the power and phase array should match the number of ports

        Returns
        -------
        a_external: array
            Wave input array  at Circuit's ports

        '''
        if len(power) != len(self.port_indexes):
            raise ValueError('Length of power array does not match the number of ports of the circuit.')
        if len(phase) != len(self.port_indexes):
            raise ValueError('Length of phase array does not match the number of ports of the circuit.')
        return np.sqrt(2*np.array(power))*np.exp(1j*np.array(phase))
    
    def _b(self, a_internal):
        '''
        Wave output array at "internal" ports
        
        Parameters
        ----------
        a_internal : array
            Wave input array at internal ports
        
        Returns
        -------
        b_internal : array
            Wave output array at internal ports
            
        Note
        ----
        Wave input array at internal ports can be derived from power and 
        phase excitation at "external" ports using `_a(power, phase)` method.
        
        '''
        # return self.s @ a_internal
        return np.matmul(self.s, a_internal)
    
    def currents(self, power, phase):
        '''
        Currents at internal ports. 
        
        NB: current direction is defined as positive when entering a node.
        
        NB: external current sign are opposite than corresponding internal ones,
            as the internal currents are actually flowing into the "port" networks

        Parameters
        ----------
        power : list or array
            Input power at external ports in Watts [W]
        phase : list or array
            Input phase at external ports in radian [rad]

        Returns
        -------
        I : complex array of shape (nfreqs, nports)
            Currents in Amperes [A] (peak) at internal ports.

        '''
        # It is possible with Circuit to define connections between 
        # multiple (>2) ports at the same time in the connection setup, like : 
        # cnx = [
        #       [(ntw1, portA), (ntw2, portB), (ntw3, portC)], ...
        #]
        # Such a case is not supported with the present calculation method
        # which only works with pair connections between ports, ie like:
        # cnx = [
        #       [(ntw1, portA), (ntw2, portB)],
        #       [(ntw2, portD), (ntw3, portC)], ...
        #]
        # It should not be a huge limitation (?), since it should be always possible
        # to add the proper splitting Network (such a "T" or hybrid or more)
        # and connect this splitting Network ports to other Network ports.
        # ie going from:
        # [ntwA]  ---- [ntwB]
        #          |
        #          |
        #        [ntwC]
        # to:
        # [ntwA] ------ [ntwD] ------ [ntwB]
        #                 |
        #                 |
        #              [ntwC]
        for inter_idx, inter in self.intersections_dict.items():
            if len(inter) > 2:
                raise NotImplementedError('Connections between more than 2 ports are not supported (yet?)')

        a = self._a(self._a_external(power, phase))
        b = self._b(a)
        z0s = self.z0
        directions = self._currents_directions
        Is = (b[:,directions[:,0]] - b[:,directions[:,1]])/np.sqrt(z0s)
        return Is


    def voltages(self, power, phase):
        '''
        Voltages at internal ports. 
        
        NB: 

        Parameters
        ----------
        power : list or array
            Input power at external ports in Watts [W]
        phase : list or array
            Input phase at external ports in radian [rad]

        Returns
        -------
        V : complex array of shape (nfreqs, nports)
            Voltages in Amperes [A] (peak) at internal ports.

        '''
        # cf currents() for more details
        for inter_idx, inter in self.intersections_dict.items():
            if len(inter) > 2:
                raise NotImplementedError('Connections between more than 2 ports are not supported (yet?)')

        a = self._a(self._a_external(power, phase))
        b = self._b(a)
        z0s = self.z0
        directions = self._currents_directions
        Vs = (b[:,directions[:,0]] + b[:,directions[:,1]])*np.sqrt(z0s)
        return Vs        
    
    def currents_external(self, power, phase):
        '''
        Currents at external ports. 
        
        NB: current direction is defined positive when "entering" into port.

        Parameters
        ----------
        power : list or array
            Input power at external ports in Watts [W]
        phase : list or array
            Input phase at external ports in radian [rad]

        Returns
        -------
        I : complex array of shape (nfreqs, nports)
            Currents in Amperes [A] (peak) at external ports. 

        '''
        a = self._a(self._a_external(power, phase))
        b = self._b(a)
        z0s = self.z0
        Is = []
        for port_idx in self.port_indexes:
            Is.append((a[port_idx] - b[:,port_idx])/np.sqrt(z0s[:,port_idx]))
        return np.array(Is).T
    
    def voltages_external(self, power, phase):
        '''
        Voltages at external ports

        Parameters
        ----------
        power : list or array
            Input power at external ports in Watts [W]
        phase : list or array
            Input phase at external ports in radian [rad]

        Returns
        -------
        V : complex array of shape (nfreqs, nports)
            Voltages in Volt [V] (peak)  at ports

        '''        
        a = self._a(self._a_external(power, phase))
        b = self._b(a)
        z0s = self.z0
        Vs = []
        for port_idx in self.port_indexes:
            Vs.append((a[port_idx] + b[:,port_idx])*np.sqrt(z0s[:,port_idx]))
        return np.array(Vs).T





'''
.. module:: skrf.media.mline
========================================
MLine (:mod:`skrf.media.MLine`)
========================================

Microstripline class


This class was made from the technical documentation [#]_ provided
by the qucs project [#]_ .
The variables  and properties of this class are coincident with
their derivations.

In addition, Djordjevic/Svensson widebande debye dielectric model is considered
to provide more realistic modelling of broadband microstrip as well as causal
time domain response.

* Quasi-static characteristic impedance model:
    In the case another model is used for effective dielectric constant,
    Hammerstad and Jensen method is used for the impedance.
    + Kirschning and Jansen
    + Hammerstad and Jensen
    + None
* Quasi-static effective dielectric constant:
    + Kirschning and Jansen
    + Hammerstad and Jensen
    + Yamashita
    + Kobayashi
    + None
* Strip thickness correction:
    + Hammerstad and Jensen, add a certain amount to W if T > 0.

.. [#] http://qucs.sourceforge.net/docs/technical.pdf
.. [#] http://www.qucs.sourceforge.net/
.. C. Svensson, G.E. Dermer,
    Time domain modeling of lossy interconnects,
    IEEE Trans. on Advanced Packaging, May 2001, N2, Vol. 24, pp.191-196.
.. Djordjevic, R.M. Biljic, V.D. Likar-Smiljanic, T.K. Sarkar,
    Wideband frequency-domain characterization of FR-4 and time-domain causality,
    IEEE Trans. on EMC, vol. 43, N4, 2001, p. 662-667.

.. autosummary::
   :toctree: generated/

   MLine

'''
import numpy as npy
from scipy.constants import  epsilon_0, mu_0, c
from numpy import real, imag, pi, sqrt,log, exp, log10, zeros, ones, tanh, \
    cosh, arctan, absolute
from .media import Media
from ..tlineFunctions import skin_depth, surface_resistivity

class MLine(Media):
    '''
    Microstripline initializer

    Parameters
    -------------
    frequency : :class:`~skrf.frequency.Frequency` object
        frequency band of the media
    z0 : number, array-like, or None
        the port impedance for media. Only needed if  its different
        from the characterisitc impedance of the transmission
    w : number, or array-like
        width of conductor, in m.
    h : number, or array-like
        height of subtrate between ground plane and conductor, in m.
    t : number, or array-like, optional
        conductor thickness, in m.
    ep_r : number, or array-like
        relative permittivity of dielectric at frequency f_epr_tand
    mu_r : number, or array-like
        relative permeability of dielectric (assumed frequency invariant)
    diel : number, or array-like
        dielectric dispersion model: djordjevicsvensson or frequencyinvariant.
    rho: number, or array-like, optional
        resistivity of conductor (None)
    tand: number, or array-like
        dielectric loss factor at frequency f_epr_tand
    rough: number, or array-like
        RMS rhougness of conductor in m.
    disp: number, or array-like
        microstripline dispersion model in
        * kirschningjansen
        * hammerstadjensen
        * yamashita
        * kobayashi
        * none
    f_low: number, or array-like
        lower frequency for wideband Debye Djordjevic/Svensson dielectric model
    f_high: number, or array-like
        higher frequency for wideband Debye Djordjevic/Svensson dielectric model
    f_epr_tand: number, or array-like
        measurement frequency for ep_r and tand of dielectric
    '''
    def __init__(self, frequency=None, z0=None, w=3, h=1.6, t=None,
                 ep_r=4.5, mu_r=1, diel='djordjevicsvensson',
                 rho=1.68e-8, tand=0, rough=0.15e-6, disp='kirschningjansen',
                 f_low=1e3, f_high=1e12, f_epr_tand=1e9,
                 *args, **kwargs):
        Media.__init__(self, frequency=frequency,z0=z0)
        
        self.w, self.h, self.t = w, h, t
        self.ep_r, self.mu_r, self.diel = ep_r, mu_r, diel
        self.rho, self.tand, self.rough, self.disp =  rho, tand, rough, disp
        self.f_low, self.f_high, self.f_epr_tand = f_low, f_high, f_epr_tand


    def __str__(self):
        f=self.frequency
        output =  \
                'Microstripline Media.  %i-%i %s.  %i points'%\
                (f.f_scaled[0],f.f_scaled[-1],f.unit, f.npoints) + \
                '\n W= %.2em, H= %.2em'% \
                (self.w,self.h)
        return output

    def __repr__(self):
        return self.__str__()

    @property
    def delta_w1(self):
        '''
        intermediary parameter. see qucs docs on microstrip lines.
        '''
        w, h, t = self.w, self.h, self.t
        if t > 0.:
            # Qucs formula 11.22 is wrong, normalized w has to be used instead (see Hammerstad and Jensen Article)
            return t/pi * log(1. + 4*exp(1.)*tanh(sqrt(6.517*w/h))**2/t) * h
        else:
            return 0.
    
    @property
    def delta_wr(self):
        '''
        intermediary parameter. see qucs docs on microstrip lines.
        '''
        delta_w1, ep_r = self.delta_w1, real(self.ep_r_f)
        if self.t > 0.:
            return delta_w1/2 * (1+1/cosh(sqrt(ep_r-1)))
        else:
            return 0.
    
    @property
    def ep_r_f(self):
        '''
        Frequency dependant relative permittivity of dielectric
        '''
        ep_r, tand  = self.ep_r, self.tand
        f_low, f_high, f_epr_tand = self.f_low, self.f_high, self.f_epr_tand
        f = self.frequency.f
        if self.diel == 'djordjevicsvensson':
            # compute the slope for a log frequency scale, tanD dependant.
            m = (ep_r*tand)  * (pi/(2*log(10)))
            # value for frequency above f_high
            ep_inf = (ep_r - 1j*ep_r*tand - m*log((f_high + 1j*f_epr_tand)/(f_low + 1j*f_epr_tand)))
            return ep_inf + m*log((f_high + 1j*f)/(f_low + 1j*f))
        elif self.diel == 'frequencyinvariant':
            return ones(self.frequency.f.shape) * (ep_r - 1j*ep_r*tand)
        else:
            raise ValueError('Unknown dielectric dispersion model')
    
    @property
    def tand_f(self):
        '''
        Frequency dependant dielectric loss factor
        '''
        ep_r = self.ep_r_f
        return -imag(ep_r) / real(ep_r)
    
    @property
    def ep_reff(self):
        '''
        Quasistatic effective relative permittivity of dielectric.
        Accounting for the filling factor between air and substrate.
        '''
        h, ep_r  = self.h, self.ep_r_f
        w1 = self.w + self.delta_w1
        wr = self.w + self.delta_wr
        return ep_re(wr, h, ep_r) * (ZL1(w1, h)/ZL1(wr, h))**2
    
    @property
    def G(self):
        '''
        intermediary parameter. see qucs docs on microstrip lines.
        '''
        ep_r, ep_reff, ZL = self.ep_r_f, self.ep_reff, self.Z0
        ZF0 = npy.sqrt(mu_0/epsilon_0)
        return (pi**2)/12 * (ep_r-1)/ep_reff * sqrt(2*pi*ZL/ZF0)
    
    @property
    def ep_reff_f(self):
        '''
        Frequency dependant effective relative permittivity of dielectric,
        accounting for microstripline dispersion.
        '''
        ep_r, ep_reff  = self.ep_r_f, self.ep_reff
        w, h = self.w + self.delta_wr, self.h
        f = self.frequency.f
        if self.disp == 'hammerstadjensen':
            ZL, G = self.Z0, self.G
            fp = ZL/(2*mu_0*h)
            return ep_r - (ep_r - ep_reff) / (1+G*(f/fp)**2)
        elif self.disp == 'kirschningjansen':
            fn = self.frequency.f * h * 1e-6
            P1 = 0.27488+(0.6315+0.525/(1+0.0157*fn)**20)*w/h -0.065683*exp(-8.7513*w/h)
            P2 = 0.33622*(1-exp(-0.03442*ep_r))
            P3 = 0.0363*exp(-4.6*w/h)*(1-exp(-(fn/38.7)**4.97))
            P4 = 1+2.751*(1-exp(-(ep_r/15.916)**8))
            Pf = P1*P2*((0.1844+P3*P4)*fn)**1.5763
            return ep_r - (ep_r-ep_reff)/(1+Pf)
        elif self.disp == 'yamashita':
            k = sqrt(ep_r/ep_reff)
            F = 4*h*f*sqrt(ep_r-1)/c * (0.5+(1+2*log(1+w/h))**2)
            return ep_reff * ((1+1/4*k*F**1.5)/(1+1/4*F**1.5))**2
        elif self.disp == 'kobayashi':
            f50 = c/(2*pi*h*(0.75+(0.75-0.332/(ep_r**1.73)*w/h))) \
                * arctan(ep_r*sqrt((ep_reff-1)/(ep_r-ep_reff)))/ \
                sqrt(ep_r-ep_reff)
            m0 = 1+1/(1+sqrt(w/h))+0.32*(1/(1+sqrt(w/h)))**3
            mc = 1+1.4/(1+w/h)*(0.15-0.235*exp(-0.45*f/f50))
            if(w/h >= 0.7):
                mc = 1
            m = m0*mc
            return ep_r - (ep_r-ep_reff)/(1+f/f50)**m
        elif self.disp == 'none':
            return ones(self.frequency.f.shape)*self.ep_reff
        else:
            raise ValueError('Unknown microstripline dispersion model')

    @property
    def Z0(self):
        '''
        Quasistatic characteristic impedance
        '''
        h, ep_reff = self.h, real(self.ep_reff)
        wr = self.w + self.delta_wr
        return ZL1(wr, h)/sqrt(ep_reff)
    
    @property
    def Z0_f(self):
        '''
        Frequency dependant characteristic impedance
        '''
        ZL, ep_reff, ep_reff_f = self.Z0, real(self. ep_reff), real(self.ep_reff_f)
        wr, h = self.w + self.delta_wr, self.h
        if self.disp == 'hammerstadjensen':
            return ZL * sqrt(ep_reff/ ep_reff_f) * (ep_reff_f-1)/(ep_reff-1)
        elif self.disp == 'kirschningjansen':
            u = wr/h
            fn = self.frequency.f * self.h * 1e-6
            ep_r = real(self.ep_r_f)
            R1 = 0.03891*ep_r**1.4
            R2 = 0.267*u**7
            R3 = 4.766*exp(-3.228*u**0.641)
            R4 = 0.016+(0.0514*ep_r)**4.524
            R5 = (fn/28.843)**12
            R6 = 22.20*u**1.92
            R7 = 1.206-0.3144*exp(-R1)*(1-exp(-R2))
            R8 = 1+1.275*(1-exp(-0.004625*R3*ep_r**1.674)*(fn/18.365)**2.745)
            R9 = 5.086*R4*R5/(0.3838+0.386*R4)*exp(-R6)/(1+1.2992*R5)*(ep_r-1)**6/(1+10*(ep_r-1)**6)
            R10 = 0.00044*ep_r**2.136+0.0184
            R11 = (fn/19.47)**6/(1+0.0962*(fn/19.47)**6)
            R12 = 1/(1+0.00245*u**2)
            R13 = 0.9408*ep_reff_f**R8-0.9603
            R14 = (0.9408-R9)*ep_reff**R8-0.9603
            R15 = 0.707*R10*(fn/12.3)**1.097
            R16 = 1+0.0503*ep_r**2*R11*(1-exp(-(u/15)**6))
            R17 = R7 * (1-1.1241*R12/R16*exp(-0.026*fn**1.15656-R15))
            return ZL * (R13/R14)**R17
        elif self.disp == 'yamashita':
            return ZL * npy.sqrt(ep_reff/ ep_reff_f) * (ep_reff_f-1)/(ep_reff-1)
        elif self.disp == 'kobayashi':
            return ZL * npy.sqrt(ep_reff/ ep_reff_f) * (ep_reff_f-1)/(ep_reff-1)
        elif self.disp == 'none':
            return npy.ones(self.frequency.f.shape)*self.Z0
        else:
            raise ValueError('Unknown microstripline dispersion model')

    @property
    def alpha_conductor(self):
        '''
        Losses due to conductor resistivity

        Returns
        --------
        alpha_conductor : array-like
                lossyness due to conductor losses
        See Also
        ----------
        surface_resistivity : calculates surface resistivity
        '''
        if self.rho is None or self.t is None:
            raise(AttributeError('must provide values conductivity to calculate this. see initializer help'))
        else:
            f = self.frequency.f
            ZF0 = npy.sqrt(mu_0/epsilon_0)
            ZL, rho, mu_r  = real(self.Z0_f), self.rho, self.mu_r
            w = self.w + self.delta_wr
            rough = self.rough
            Kr  = 1 + 2/pi*arctan(1.4*(rough/skin_depth(f, rho, mu_r))**2)
            Ki  = exp(-1.2*(ZL/ZF0)**0.7)
            Rs  = surface_resistivity(f=f, rho=rho, mu_r=1)
        return  Rs*Kr*Ki/(ZL*w)
    
    @property
    def alpha_dielectric(self):
        '''
        Losses due to dielectric

        '''
        ep_r, ep_reff, tand = real(self.ep_r_f), real(self.ep_reff), self.tand_f
        f = self.frequency.f
        return pi*f/c * ep_r/sqrt(ep_reff) * (ep_reff-1)/(ep_r-1) * tand
    
    @property
    def beta_phase(self):
        '''
        Phase parameter
        '''
        ep_reff, f = real(self.ep_reff_f), self.frequency.f
        return 2*pi*f*sqrt(ep_reff)/c

    @property
    def gamma(self):
        '''
        Propagation constant


        See Also
        --------
        alpha_conductor : calculates losses to conductor
        alpha_dielectric: calculates losses to dielectric
        beta            : calculates phase parameter
        '''
        f = self.frequency.f
        alpha = zeros(len(f))
        beta  = zeros(len(f))
        beta  = self.beta_phase
        alpha = self.alpha_dielectric
        if self.rho is not None:
            alpha += self.alpha_conductor
        return alpha + 1j*beta

def a(u):
    '''
    intermediary parameter. see qucs docs on microstrip lines.
    '''
    return 1. + 1./49.*log((u**4+(u/52)**2)/(u**4+0.432)) + 1./18.7*log(1+(u/18.1)**3)
    
def b(ep_r):
    '''
    intermediary parameter. see qucs docs on microstrip lines.
    '''
    return 0.564 * ((ep_r-0.9)/(ep_r+3.))**0.053
def f(u):
    '''
    intermediary parameter. see qucs docs on microstrip lines.
    '''
    return 6 + (2*pi-6)*exp(-(30.666/u)**0.7528)
    
def ZL1(w, h):
    '''
    intermediary parameter. see qucs docs on microstrip lines.
    '''
    u = w/h
    ZF0 = sqrt(mu_0/epsilon_0)
    return ZF0/(2*pi)*log(f(u)/u + sqrt(1.+(2./u)**2))

def ep_re(w, h, ep_r):
    '''
    intermediary parameter. see qucs docs on microstrip lines.
    '''
    u = w/h
    return (ep_r+1)/2 + (ep_r-1)/2 * (1.+10./u)**(-a(u)*b(ep_r))



# -*- coding: utf-8 -*-
import unittest
import os
import skrf as rf
from skrf.media import Coaxial
import numpy as npy
from numpy.testing import assert_almost_equal, assert_array_almost_equal, run_module_suite

class MediaTestCase(unittest.TestCase):
    '''

    '''
    def setUp(self):
        '''

        '''
        self.files_dir = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            'qucs_prj'
            )



    def test_line(self):
        '''
        '''
        fname = os.path.join(self.files_dir,\
                'coaxial.s2p')
        qucs_ntwk = rf.Network(fname)

        a_media = Coaxial(
            frequency = qucs_ntwk.frequency,
            Dint=1e-3, Dout=3e-3, epsilon_r=2.29, \
            tan_delta=4e-4, sigma=1./1.68e-8 \
            )
        skrf_ntwk = a_media.thru(z0=50)**a_media.line(200e-3,'m')\
                    **a_media.thru(z0=50)
        # Equal assertion fails if tan_delta or resistivity are non-zero
        #self.assertEqual(qucs_ntwk, skrf_ntwk)
        self.assertTrue(
            max(abs(skrf_ntwk.s_mag[:,1,0] - qucs_ntwk.s_mag[:,1,0])) < 1e-3
            )


    def test_init_from_attenuation_VF_units(self):
        """
        Test the attenuation unit conversions in the Coaxial classmethod
        `from_attenuation_VF_units`.
        """
        # create a dummy Coaxial media for various attenuation and test the
        # resulting alpha values (real part of gamma)
        frequency = rf.Frequency(npy.random.rand(), unit='GHz', npoints=1)
        _att = npy.random.rand()
        # dB/m
        coax = Coaxial.from_attenuation_VF(frequency=frequency, VF=1, att=_att, unit='dB/m')
        assert_almost_equal(coax.gamma.real,  rf.db_2_np(_att))
        # dB/100m
        coax = Coaxial.from_attenuation_VF(frequency=frequency, VF=1, att=_att, unit='dB/100m')
        assert_almost_equal(coax.gamma.real,  rf.db_2_np(_att)/100)
        # dB/feet
        coax = Coaxial.from_attenuation_VF(frequency=frequency, VF=1, att=_att, unit='dB/feet')
        assert_almost_equal(coax.gamma.real,  rf.db_2_np(_att)*rf.meter_2_feet())
        # dB/100m
        coax = Coaxial.from_attenuation_VF(frequency=frequency, VF=1, att=_att, unit='dB/100feet')
        assert_almost_equal(coax.gamma.real,  rf.db_2_np(_att)/100*rf.meter_2_feet())
        # Neper/m
        coax = Coaxial.from_attenuation_VF(frequency=frequency, VF=1, att=_att, unit='Np/m')
        assert_almost_equal(coax.gamma.real,  _att)
        # Neper/feet
        coax = Coaxial.from_attenuation_VF(frequency=frequency, VF=1, att=_att, unit='Np/feet')
        assert_almost_equal(coax.gamma.real,  _att*rf.meter_2_feet())


    def test_init_from_attenuation_VF_array_att(self):
        """
        Test passing array as attenuation in the Coaxial classmethod
        `from_attenuation_VF_units`.
        """
        # create a Coaxial media for frequency-dependant attenuation and 
        # test the resulting alpha values (real part of gamma)
        frequency = rf.Frequency(start=1, stop=2, unit='GHz', npoints=101)
        # k0k1k2 attenuation model
        # values taken for HUBER+SUHNER DATA SHEET Coaxial Cable S_10172_B-1
        # attenuation in dB/m for frequency in GHz
        att = 0 + 0.0826*npy.sqrt(frequency.f_scaled) + 0.0129*frequency.f_scaled

        coax = Coaxial.from_attenuation_VF(frequency=frequency, att=att, unit='dB/m')
        # check alpha in gamma
        assert_array_almost_equal(rf.db_2_np(att), coax.gamma.real)

        # if the attenuation array length does not match the frequency,
        # should raise a ValueError
        frequency2 = rf.Frequency(start=1, stop=2, unit='GHz', npoints=10)
        with self.assertRaises(ValueError):
            coax = Coaxial.from_attenuation_VF(frequency=frequency2, att=att)


if __name__ == "__main__":
    # Launch all tests
    run_module_suite()




'''
.. module:: skrf.media.definedAEpTandZ0

======================================================
DefinedAEpTandZ0 (:mod:`skrf.media.definedAEpTandZ0`)
======================================================

Transmission line medium defined by A, Ep, Tand and Z0.

This medium is defined by attenuation A, relative permittivity Ep_r,
loss angle tand and characteristic impedance Z0.

Djirdjevic/Svennson dispersion model is provided for dielectric, default
behaviour is frequency invariant.

.. autosummary::
   :toctree: generated/

   DefinedAEpTandZ0
'''
from scipy.constants import  epsilon_0, c
from numpy import real, imag, sqrt, ones, zeros, pi, log
from .media import Media

class DefinedAEpTandZ0(Media):
    '''
    Transmission line medium defined by A, Ep, Tand and Z0.
    
    This medium is defined by attenuation `A`, relative permittivity `Ep_r`,
    loss angle `tand` and characteristic impedance `Z0`.
    
    Djirdjevic/Svennson dispersion model is provided for dielectric, default
    behaviour is frequency invariant.
    
    A DefinedAEpTandZ0 medium is contructed: 
     * from scalar attenuation `A`, relative permittivity `Ep_r`,
       loss angle `tand` and characteristic impedance `Z0`.
       Frequency invariant behaviour or Djirdjevic/Svennson dispersion can be
       choosed trough `model`. Default is frequency invariant.
     
     See Examples.

    Parameters
    -----------
    frequency : :class:`~skrf.frequency.Frequency` object
        Frequency band of this transmission line medium
    z0 : number, array-like, or None
        The port impedance for this medium. Only needed if different from the
        characterisitc impedance of the transmission line.
        If z0 is None then will default to Z0
    A : number, array-like, default 0.0
        Attenuation due to conductor loss in dB/m/sqrt(Hz)
        The attenuation :math:`A(f)`at frequency :math:`f` is:
        .. math::
            A(f) = A*\sqrt{\frac{f}/{f_A}}
    f_A: number, default 1.0
        Frequency scaling in Hz for the attenuation. See A.
    ep_r : number, array-like, default 1.0
        Real part of the relative permittivity of the dielectric.
        if `model='frequencyinvariant'`, the complex relative permittivity is:
        .. math::
            \epsilon_r(f) = ep\_r + j \cdot ep\_r \cdot tanD
        if `model='djordjevicsvensson'`, the complex relative permittivity is:
        .. math::
            \epsilon_r(f) = ep_\inf + m \cdot \log{\frac{f_{high} - f_0}{f_{low} - f_\epsilon}}
        In this case, the value of ep_r and tanD are given for frequency f_ep:
        .. math::
            \epsilon_r(f_\epsilon) = ep\_r + j \cdot ep\_r \cdot tanD
    tanD: number, array-like, default 0.0
        Dielectric relative permittivity loss tangent. See ep_r.
    Z0: number, array-like, default 50.0
        Quasi-static characteristic impedance of the medium.
    f_low: number, default 1e3, , optional
        Low frequency in Hz for  for Djirdjevic/Svennson dispersion model.
        See ep_r.
    f_high: number, default 1e12, optional
        High frequency in Hz for  for Djirdjevic/Svennson dispersion model.
        See ep_r.
    f_ep: number, default 1e9, , optional
        Specification frequency in Hz for  for Djirdjevic/Svennson dispersion model.
        ep_r and tanD parameters are specified for this frequency. See ep_r.
    model: string, 'frequencyinvariant' or 'djordjevicsvensson', optional
        Use Djirdjevic/Svennson wideband Debye dispersion model or not.
        Default is frequency invariant behaviour.
    \*args, \*\*kwargs : arguments and keyword arguments


    Examples
    -----------
    >>>from skrf.media.definedAEpTandZ0 import DefinedAEpTandZ0
    >>>from skrf.frequency import Frequency
    >>>f = Frequency(75,110,101,'ghz')
    >>>DefinedAEpTandZ0(frequency=f, A=1, f_A=1e9, ep_r=3, Z0=50)   
    >>>DefinedAEpTandZ0(frequency=f, A=1, f_A=1e9, ep_r=3, tand=0.02, Z0=50)
    >>>DefinedAEpTandZ0(frequency=f, A=1, f_A=1e9, ep_r=3, tand=0.02, Z0=50,
                        f_low=1e3, f_high=1e12, f_Ep=1e9,
                        model='djordjevicsvensson')
    
    
    '''
    def __init__(self, frequency=None, z0=None,
                 A=0.0, f_A=1.0, ep_r=1.0, tanD=0.0, Z0=50.0,
                 f_low=1.0e3, f_high=1.0e12, f_ep=1.0e9,
                 model='frequencyinvariant', *args, **kwargs):
        
        Media.__init__(self, frequency=frequency,z0=z0)
        self.A, self.f_A = A, f_A
        self.ep_r, self.tanD = ep_r, tanD
        self._Z0 = Z0
        self.f_low, self.f_high, self.f_ep = f_low, f_high, f_ep
        self.model = model
    
    def __str__(self):
        f=self.frequency
        output = 'DefinedAEpTandZ0 medium.  %i-%i %s.  %i points'%\
                (f.f_scaled[0],f.f_scaled[-1],f.unit, f.npoints)
        return output

    def __repr__(self):
        return self.__str__()
    
    @property
    def ep_r_f(self):
        '''
        Frequency dependant complex relative permittivity of dielectric
        '''
        ep_r, tand  = self.ep_r, self.tanD
        f_low, f_high, f_ep = self.f_low, self.f_high, self.f_ep
        f = self.frequency.f
        if self.model == 'djordjevicsvensson':
            # compute the slope for a log frequency scale, tanD dependant.
            m = (ep_r*tand)  * (pi/(2*log(10)))
            # value for frequency above f_high
            ep_inf = (ep_r - 1j*ep_r*tand - m*log((f_high + 1j*f_ep)/(f_low + 1j*f_ep)))
            return ep_inf + m*log((f_high + 1j*f)/(f_low + 1j*f))
        elif self.model == 'frequencyinvariant':
            return ones(self.frequency.f.shape) * (ep_r - 1j*ep_r*tand)
        else:
            raise ValueError('Unknown dielectric dispersion model')
    
    @property
    def tand_f(self):
        '''
        Frequency dependant dielectric loss factor
        '''
        ep_r = self.ep_r_f
        return -imag(ep_r) / real(ep_r)
         
    @property
    def alpha_conductor(self):
        '''
        Losses due to conductor resistivity
        '''
        A, f_A, f = self.A, self.f_A, self.frequency.f
        return A * log(10)/20 * sqrt(f/f_A)
    
    @property
    def alpha_dielectric(self):
        '''
        Losses due to dielectric
        '''
        ep_r, tand = real(self.ep_r_f), self.tand_f
        f = self.frequency.f
        return pi*sqrt(ep_r)*f/c * tand
    
    @property
    def beta_phase(self):
        '''
        Phase parameter
        '''
        ep_r, f = real(self.ep_r_f), self.frequency.f
        return 2*pi*f*sqrt(ep_r)/c

    @property
    def gamma(self):
        '''
        Propagation constant

        See Also
        --------
        alpha_conductor : calculates conductor losses
        alpha_dielectric: calculates dielectric losses
        beta_phase      : calculates phase parameter
        '''
        beta  = self.beta_phase
        alpha = self.alpha_conductor + self.alpha_dielectric
        return alpha + 1j*beta
    
    @property
    def Z0(self):
        return self._Z0
    
    @Z0.setter
    def Z0(self, val):
        self._Z0 = val





'''
.. module:: skrf.media.distributedCircuit

============================================================
distributedCircuit (:mod:`skrf.media.distributedCircuit`)
============================================================

A transmission line mode defined in terms of distributed impedance and admittance values. 

.. autosummary::
   :toctree: generated/

   DistributedCircuit

'''

from copy import deepcopy
from scipy.constants import  epsilon_0, mu_0, c,pi, mil
import numpy as npy
from numpy import sqrt, exp, array,tan,sin,cos,inf, log, real,imag,\
         interp, linspace, shape,zeros, reshape

from ..tlineFunctions import electrical_length
from .media import Media, DefinedGammaZ0


from ..constants import INF, ONE, ZERO

class DistributedCircuit(Media):
    '''
    A transmission line mode defined in terms of distributed impedance and admittance values. 

    Parameters
    ------------
    frequency : :class:`~skrf.frequency.Frequency` object
        frequency band of the media
    z0 : number, array-like, or None
        the port impedance for media. Only needed if  its different
        from the characterisitc impedance of the transmission
        line. if z0 is None then will default to Z0
    C : number, or array-like
            distributed capacitance, in F/m
    L : number, or array-like
            distributed inductance, in  H/m
    R : number, or array-like
            distributed resistance, in Ohm/m
    G : number, or array-like
            distributed conductance, in S/m


    Notes
    ----------
    if C,I,R,G are vectors they should be the same length
    
    :class:`DistributedCircuit` is `Media` object representing a 
    transmission line mode defined in terms of  distributed impedance
    and admittance values. 

    A Distributed Circuit may be defined in terms
    of the following attributes,

    ================================  ================  ================
    Quantity                          Symbol            Property
    ================================  ================  ================
    Distributed Capacitance           :math:`C^{'}`     :attr:`C`
    Distributed Inductance            :math:`L^{'}`     :attr:`L`
    Distributed Resistance            :math:`R^{'}`     :attr:`R`
    Distributed Conductance           :math:`G^{'}`     :attr:`G`
    ================================  ================  ================


    The following quantities may be calculated, which are functions of 
    angular frequency (:math:`\omega`):

    ===================================  ==================================================  ==============================
    Quantity                             Symbol                                              Property
    ===================================  ==================================================  ==============================
    Distributed Impedance                :math:`Z^{'} = R^{'} + j \\omega L^{'}`              :attr:`Z`
    Distributed Admittance               :math:`Y^{'} = G^{'} + j \\omega C^{'}`              :attr:`Y`
    ===================================  ==================================================  ==============================


    The properties which define their wave behavior:

    ===================================  ============================================  ==============================
    Quantity                             Symbol                                        Method
    ===================================  ============================================  ==============================
    Characteristic Impedance             :math:`Z_0 = \\sqrt{ \\frac{Z^{'}}{Y^{'}}}`     :func:`Z0`
    Propagation Constant                 :math:`\\gamma = \\sqrt{ Z^{'}  Y^{'}}`         :func:`gamma`
    ===================================  ============================================  ==============================

    Given the following definitions, the components of propagation
    constant are interpreted as follows:

    .. math::
        +\\Re e\\{\\gamma\\} = \\text{attenuation}

        -\\Im m\\{\\gamma\\} = \\text{forward propagation}
    

    See Also 
    --------
    from_media

    '''
    
    def __init__(self, frequency=None, z0=None, C=90e-12, L=280e-9, R=0, G=0,
                *args, **kwargs):
        super(DistributedCircuit, self).__init__(frequency=frequency, 
                                                 z0=z0)
        self.C, self.L, self.R, self.G = C,L,R,G


    def __str__(self):
        f=self.frequency
        try:
            output =  \
                'Distributed Circuit Media.  %i-%i %s.  %i points'%\
                (f.f_scaled[0],f.f_scaled[-1],f.unit, f.npoints) + \
                '\nL\'= %.2f, C\'= %.2f,R\'= %.2f, G\'= %.2f, '% \
                (self.L, self.C,self.R, self.G)
        except(TypeError):
            output =  \
                'Distributed Circuit Media.  %i-%i %s.  %i points'%\
                (f.f_scaled[0],f.f_scaled[-1],f.unit, f.npoints) + \
                '\nL\'= %.2f.., C\'= %.2f..,R\'= %.2f.., G\'= %.2f.., '% \
                (self.L[0], self.C[0],self.R[0], self.G[0])
        return output

    def __repr__(self):
        return self.__str__()

    @classmethod
    def from_media(cls, my_media, *args, **kwargs):
        '''
        Initializes a DistributedCircuit from an existing
        :class:'~skrf.media.media.Media' instance.

        Parameters
        ------------
        my_media : :class:'~skrf.media.media.Media' instance.
            the media object
        '''

        w  =  my_media.frequency.w
        gamma = my_media.gamma
        Z0 = my_media.Z0
        z0 = my_media.z0

        Y = gamma/Z0
        Z = gamma*Z0
        G,C = real(Y), imag(Y)/w
        R,L = real(Z), imag(Z)/w
        return cls(frequency = my_media.frequency, 
                   z0 = z0, C=C, L=L, R=R, G=G, *args, **kwargs)
    
    @classmethod
    def from_csv(self, *args, **kw):
        d = DefinedGammaZ0.from_csv(*args,**kw)
        return self.from_media(d)

    @property
    def Z(self):
        '''
        Distributed Impedance, :math:`Z^{'}`

        Defined as

        .. math::
                Z^{'} = R^{'} + j \\omega L^{'}


        Returns
        --------
        Z : numpy.ndarray
                Distributed impedance in units of ohm/m
        '''
        w  = self.frequency.w
        return self.R + 1j*w*self.L

    @property
    def Y(self):
        '''
        Distributed Admittance, :math:`Y^{'}`

        Defined as

        .. math::
                Y^{'} = G^{'} + j \\omega C^{'}

        Returns
        --------
        Y : numpy.ndarray
                Distributed Admittance in units of S/m
        '''

        w  = self.frequency.w
        return self.G + 1j*w*self.C

    @property
    def Z0(self):
        '''
        Characteristic Impedance, :math:`Z0`

        .. math::
                Z_0 = \\sqrt{ \\frac{Z^{'}}{Y^{'}}}

        Returns
        --------
        Z0 : numpy.ndarray
                Characteristic Impedance in units of ohms
        '''

        return sqrt(self.Z/self.Y)

    @property
    def gamma(self):
        '''
        Propagation Constant, :math:`\\gamma`

        Defined as,

        .. math::
                \\gamma =  \\sqrt{ Z^{'}  Y^{'}}

        Returns
        --------
        gamma : numpy.ndarray
                Propagation Constant,

        Notes
        ---------
        The components of propagation constant are interpreted as follows:

        positive real(gamma) = attenuation
        positive imag(gamma) = forward propagation
        '''
        return sqrt(self.Z*self.Y)





'''
.. module:: skrf.media.cpw

========================================
cpw (:mod:`skrf.media.cpw`)
========================================

Coplanar waveguide class

This class was made from the technical documentation [#]_ provided
by the qucs project [#]_ .
The variables  and properties of this class are coincident with
their derivations.

.. [#] http://qucs.sourceforge.net/docs/technical.pdf
.. [#] http://www.qucs.sourceforge.net/

.. autosummary::
   :toctree: generated/

   CPW

'''
from scipy.constants import  epsilon_0, mu_0
from scipy.special import ellipk
from numpy import real, imag,pi,sqrt,log,zeros, ones
from .media import Media
from ..tlineFunctions import skin_depth, surface_resistivity

class CPW(Media):
    '''
    Coplanar Waveguide Media

    Parameters
    -------------
    frequency : :class:`~skrf.frequency.Frequency` object, optional
        frequency band of the media. The default is None.
    z0 : number, array-like, optional
        the port impedance for media. The default is None.
        Only needed if  its different from the characterisitc impedance 
        of the transmission.
    w : number, or array-like, optional
            width of center conductor, in m. Default is 70.
    s : number, or array-like
            width of gap, in m. Default is 4.
    ep_r : number, or array-like, optional
            relative permativity of substrate. Default is 3.
    t : number, or array-like, optional
            conductor thickness, in m. Default is None (metalization thickness neglected)
    rho: number, or array-like, optional
            resistivity of conductor. Default is None
      
    '''
    def __init__(self, frequency=None, z0=None, w=70, s=4,
                 ep_r=3, t=None, rho=None,  *args, **kwargs):
        Media.__init__(self, frequency=frequency,z0=z0)

        self.w, self.s, self.ep_r, self.t, self.rho =\
                w, s, ep_r, t, rho


    def __str__(self):
        f=self.frequency
        output =  \
                'Coplanar Waveguide Media.  %i-%i %s.  %i points'%\
                (f.f_scaled[0],f.f_scaled[-1],f.unit, f.npoints) + \
                '\n W= %.2em, S= %.2em'% \
                (self.w,self.s)
        return output

    def __repr__(self):
        return self.__str__()

    @property
    def ep_re(self):
        '''
        Effective permittivity of the CPW (also known as Keff).
        
        If the thickness of the dielectric susbtrate is large, 
        the effective dielectric constant of the even mode can be approx as:
        
        .. math::

                \epsilon_{eff} = \\frac{\epsilon_r + 1}{2} 
        
        The effective permittivity can be defined as in the case of a 
        microstripa line, that is as the square of ratio 
        of the capacitance per unit length to the phase velocity. 
        '''
        return (self.ep_r+1)/2.0

    @property
    def k1(self):
        '''
        Intermediary parameter. see qucs docs on cpw lines.
        
        Defined as:
            
        .. math::
                
                k = \\frac{w}{w + 2s}
        
        '''
        return self.w/(self.w +2*self.s)

    @property
    def K_ratio(self):
        '''
        intermediary parameter. see qucs docs on cpw lines.
        K_ratio is the ratio of two elliptic integrals
        '''
        k1 = self.k1
        # k prime
        k_p = sqrt(1 - k1**2)

        if (0 <= k1 <= 1/sqrt(2)):
            return pi/log(2*(1 + sqrt(k1))/(1 - sqrt(k1)))
        elif (1/sqrt(2) < k1 <= 1):
            return (log(2*(1 + sqrt(k_p))/(1 - sqrt(k_p)) ))/pi




    @property
    def alpha_conductor(self):
        '''
        Losses due to conductor resistivity

        Returns
        --------
        alpha_conductor : array-like
                lossyness due to conductor losses
        See Also
        ----------
        surface_resistivity : calculates surface resistivity
        '''
        if self.rho is None or self.t is None:
            raise(AttributeError('must provide values conductivity and conductor thickness to calculate this. see initializer help'))

        t, k1, ep_re = self.t, self.k1,self.ep_re
        r_s = surface_resistivity(f=self.frequency.f, rho=self.rho, \
                mu_r=1)
        a = self.w/2.
        b = self.s+self.w/2.
        K = ellipk      # complete elliptical integral of first kind
        K_p = lambda x: ellipk(sqrt(1-x**2)) # ellipk's compliment

        return ((r_s * sqrt(ep_re)/(480*pi*K(k1)*K_p(k1)*(1-k1**2) ))*\
                (1./a * (pi+log((8*pi*a*(1-k1))/(t*(1+k1)))) +\
                 1./b * (pi+log((8*pi*b*(1-k1))/(t*(1+k1))))))



    @property
    def Z0(self):
        '''
        Characterisitc impedance
        '''
        return (30.*pi / sqrt(self.ep_re) * self.K_ratio)*ones(len(self.frequency.f), dtype='complex')

    @property
    def gamma(self):
        '''
        Propagation constant


        See Also
        --------
        alpha_conductor : calculates losses to conductors
        '''
        beta = 1j*2*pi*self.frequency.f*sqrt(self.ep_re*epsilon_0*mu_0)
        alpha = zeros(len(beta))
        if self.rho is not None and self.t is not None:
            alpha = self.alpha_conductor

        return beta+alpha



import unittest
import os
import numpy as npy

import skrf as rf


class FrequencyTestCase(unittest.TestCase):
    '''

    '''
    def setUp(self):
        '''
        '''
        self.test_dir = os.path.dirname(os.path.abspath(__file__))+'/'

    def test_create_linear_sweep(self):
        freq = rf.Frequency(1,10,10,'ghz')
        self.assertTrue((freq.f == npy.linspace(1,10,10)*1e9).all())
        self.assertTrue((freq.f_scaled ==npy.linspace(1,10,10)).all())

    def test_create_log_sweep(self):
        freq = rf.Frequency(1,10,10,'ghz', sweep_type='log')
        #Check end points
        self.assertTrue((freq.f[0] == 1e9))
        self.assertTrue((freq.f[-1] == 10e9))
        spacing = [freq.f[i+1]/freq.f[i] for i in range(len(freq.f)-1)]
        #Check that frequency is increasing
        self.assertTrue(all(s > 1 for s in spacing))
        #Check that ratio of adjacent frequency points is identical
        self.assertTrue(all(abs(spacing[i] - spacing[0]) < 1e-10 for i in range(len(spacing))))

    def test_create_rando_sweep(self):
        f = npy.array([1,5,200])
        freq = rf.Frequency.from_f(f,unit='khz')
        self.assertTrue((freq.f ==f*1e3).all())
        self.assertTrue((freq.f_scaled== f).all())

    def test_rando_sweep_from_touchstone(self):
        '''
        this also tests the ability to read a touchstone file.
        '''
        rando_sweep_ntwk = rf.Network(os.path.join(self.test_dir, 'ntwk_arbitrary_frequency.s2p'))
        self.assertTrue((rando_sweep_ntwk.f == \
            npy.array([1,4,10,20])).all())

    def test_slicer(self):
        a = rf.Frequency.from_f([1,2,4,5,6])

        b = a['2-5ghz']
        tinyfloat = 1e-12
        self.assertTrue((abs(b.f - [2e9,4e9,5e9]) < tinyfloat).all())

suite = unittest.TestLoader().loadTestsFromTestCase(FrequencyTestCase)
unittest.TextTestRunner(verbosity=2).run(suite)



import re

null_parameter = re.compile(",{2,}")  # detect optional null parameter as two consecutive commas, and remove
converters = {
    "str": str,
    "int": int,
    "float": float,
    "bool": lambda x: bool(int(x)),
}


def to_string(value):
    tval = type(value)
    if tval is str:
        return value
    elif tval is bool:
        return str(int(value))
    elif tval in (list, tuple):
        return ",".join(map(to_string, value))
    elif value is None:
        return ""
    else:
        return str(value)


def scpi_preprocess(command_string, *args):
    args = list(args)
    for i, arg in enumerate(args):
        args[i] = to_string(arg)
    cmd = command_string.format(*args)
    return null_parameter.sub(",", cmd)


def process_query(query, csv=False, strip_outer_quotes=True, returns="str"):
    if strip_outer_quotes is True:
        if query[0] + query[-1] in ('""', "''"):
            query = query[1:-1]
    if csv is True:
        query = query.split(",")

    converter = None if returns == "str" else converters.get(returns, None)
    if converter:
        if csv is True:
            query = list(map(converter, query))
        else:
            query = converter(query)

    return query


class SCPI(object):
    def __init__(self, resource):
        self.resource = resource
        self.echo = False  # print scpi command string to scpi out

    def write(self, scpi, *args, **kwargs):
        if self.echo:
            print(scpi)
        self.resource.write(scpi, *args, **kwargs)

    def query(self, scpi, *args, **kwargs):
        if self.echo:
            print(scpi)
        return self.resource.query(scpi, *args, **kwargs)

    def write_values(self, scpi, *args, **kwargs):
        if self.echo:
            print(scpi)
        self.resource.write_values(scpi, *args, **kwargs)

    def query_values(self, scpi, *args, **kwargs):
        if self.echo:
            print(scpi)
        return self.resource.query_values(scpi, *args, **kwargs)

    def set_active_channel(self, channel=1):
        """Selects channel as the active channel.
    
         INSTrument:NSELect <Channel>
         """
        scpi_command = scpi_preprocess(":INST:NSEL {:}", channel)
        self.write(scpi_command)

    def set_averaging_clear(self, cnum=1):
        """no help available"""
        scpi_command = scpi_preprocess(":SENS{:}:AVER:CLE", cnum)
        self.write(scpi_command)

    def set_averaging_count(self, cnum=1, FACTOR=10):
        """no help available"""
        scpi_command = scpi_preprocess(":SENS{:}:AVER:COUN {:}", cnum, FACTOR)
        self.write(scpi_command)

    def set_averaging_state(self, cnum=1, ONOFF="ON"):
        """no help available"""
        scpi_command = scpi_preprocess(":SENS{:}:AVER:STAT {:}", cnum, ONOFF)
        self.write(scpi_command)

    def set_channel_name(self, cnum=1, CNAME=""):
        """Assigns a name to channel number <Ch>.
    
         The channel must be created before (CONFigure:CHANnel<Ch>[:STATe] ON).
         Moreover it is not possible to assign the same name to two
         different channels. CONFigure:CHANnel<Ch>:CATalog? returns a list of
         all defined channels with their names.
    
         CONFigure:CHANnel<Ch>:NAME '<Ch_name>'
         """
        scpi_command = scpi_preprocess(":CONF:CHAN{:}:NAME '{:}'", cnum, CNAME)
        self.write(scpi_command)

    def set_channel_state(self, cnum=1, STATE="ON"):
        """Creates or deletes channel no. <Ch> and selects it as the active channel.
    
         CONFigure:CHANnel<Ch>:NAME defines the channel name.
         """
        scpi_command = scpi_preprocess(":CONF:CHAN{:}:STAT {:}", cnum, STATE)
        self.write(scpi_command)

    def set_channel_trace_name(self, cnum=1, TRNAME=""):
        """Assigns a (new) name to the active trace in channel <Ch>."""
        scpi_command = scpi_preprocess(":CONF:CHAN{:}:TRAC:REN '{:}'", cnum, TRNAME)
        self.write(scpi_command)

    def set_converter_mode(self, cnum=1, MODE="RILI"):
        """Selects the test setup (internal or external sources)
    
         for the frequency converter measurement (with option ZVA-K8, Converter Control).
    
         [SENSe<Ch>:]FREQuency:CONVersion:DEVice:MODE RILI | RILE | RILI4 | RILI56
    
         The available test setups depend on the instrument type:
    
                             RILI    RILE    RILI4    RILI56
         ---------------------------------------------------
         R&S ZVA 24|40|50      x      x
         R&S ZVA 67                   x        x
         R&S ZVT 20                            x        x
    
         <Ch>    Channel number. This suffix is ignored, the command affects all channels.
         RILI    RF internal, LO internal (4-Port)
                 Ports 1 and 2: Converter RF Signal
                 Ports 3 and 4: Converter LO
         RILE    RF internal, LO external (4-Port)
                 Ports 1 to 4: Converter RF Signal
                 External Generator #1: Converter LO
         RILI4    RF internal, LO internal, Port4=LO (R&S ZVA 67 only)
                 Ports 1 and 2: Converter RF Signal
                 Port 4: Converter LO
         RILI56    RF internal, LO internal, 6-Port (R&S ZVT 20 only)
                 Ports 1 to 4: Converter RF Signal
                 Ports 5 and 6: Converter LO 
        """
        scpi_command = scpi_preprocess(":SENS{:}:FREQ:CONV:DEV:MODE {:}", cnum, MODE)
        self.write(scpi_command)

    def set_converter_name(self, cnum=1, TYPE=""):
        """Selects the frequency converter type for enhanced frequency-converting measurements
    
         (with option ZVA-K8, Converter Control).
    
         The preset configuration will be reset to Instrument scope (SYSTem:PRESet:SCOPe ALL)
         and Factory Preset (SYSTem:PRESet:USER:STATe OFF).
    
         [SENSe<Ch>:]FREQuency:CONVersion:DEVice:NAME '<Converter Type>'
    
         | TYPE      | frequency range    |
         |-----------|--------------------|
         | R&S®ZC170 | 110 GHz to 170 GHz |
         | R&S®ZC220 | 140 GHz to 220 GHz |
         | R&S®ZC330 | 220 GHz to 330 GHz |
         | R&S®ZC500 | 330 GHz to 500 GHz |
         |-----------|--------------------|
     
        """
        scpi_command = scpi_preprocess(":SENS{:}:FREQ:CONV:DEV:NAME '{:}'", cnum, TYPE)
        self.write(scpi_command)

    def set_corr_connection(self, cnum=1, pnum=1, CONN=""):
        """no help available"""
        scpi_command = scpi_preprocess(":SENS{:}:CORR:COLL:CONN{:} {:}", cnum, pnum, CONN)
        self.write(scpi_command)

    def set_corr_connection_genders(self, cnum=1, pnum=1, GENDERS="ALL"):
        """no help available"""
        scpi_command = scpi_preprocess(":SENS{:}:CORR:COLL:CONN{:}:GEND {:}", cnum, pnum, GENDERS)
        self.write(scpi_command)

    def set_corr_connection_ports(self, cnum=1, pnum=1, PORTS="ALL"):
        """no help available"""
        scpi_command = scpi_preprocess(":SENS{:}:CORR:COLL:CONN{:}:PORT {:}", cnum, pnum, PORTS)
        self.write(scpi_command)

    def set_corr_state(self, cnum=1, STATE="ON"):
        """no help available"""
        scpi_command = scpi_preprocess(":SENS{:}:CORR:STAT {:}", cnum, STATE)
        self.write(scpi_command)

    def set_data(self, cnum=1, SDATA="", data=None):
        """no help available"""
        scpi_command = scpi_preprocess(":CALC{:}:DATA {:},", cnum, SDATA)
        self.write_values(scpi_command, data)

    def set_disp_color(self, COLOR="DBAC"):
        """no help available"""
        scpi_command = scpi_preprocess(":SYST:DISP:COL {:}", COLOR)
        self.write(scpi_command)

    def set_disp_maximize(self, wnum=1, ONOFF="OFF"):
        """Maximizes all diagram areas in the active setup or restores the previous display configuration."""
        scpi_command = scpi_preprocess(":DISP:WIND{:}:MAX {:}", wnum, ONOFF)
        self.write(scpi_command)

    def set_disp_name(self, wnum=1, name=""):
        """Defines a name for diagram area <Wnd>.
    
         The name appears in the list of diagram areas, to be queried by DISPlay[:WINDow<Wnd>]:CAT?.
         """
        scpi_command = scpi_preprocess(":DISP:WIND{:}:NAME {:}", wnum, name)
        self.write(scpi_command)

    def set_disp_state(self, wnum=1, ONOFF=""):
        """Creates or deletes a diagram area, identified by its area number <Wnd>.
    
         DISPlay[:WINDow<Wnd>]:STATe <Boolean>
         """
        scpi_command = scpi_preprocess(":DISP:WIND{:}:STAT {:}", wnum, ONOFF)
        self.write(scpi_command)

    def set_disp_title_data(self, wnum=1, TITLE=""):
        """Defines a title for diagram area <Wnd>.
    
         DISPlay[:WINDow<Wnd>]:TITLe:DATA '<string>'
         """
        scpi_command = scpi_preprocess(":DISP:WIND{:}:TITL:DATA {:}", wnum, TITLE)
        self.write(scpi_command)

    def set_disp_title_state(self, wnum=1, ONOFF="ON"):
        """Displays or hides the title for area number <Wnd>, defined by means of DISPlay:WINDow<Wnd>:TITLe:DATA.
    
         DISPlay[:WINDow<Wnd>]:TITLe[:STATe] <Boolean>
         """
        scpi_command = scpi_preprocess(":DISP:WIND{:}:TITL:STAT {:}", wnum, ONOFF)
        self.write(scpi_command)

    def set_disp_trace_auto(self, wnum=1, tnum=1, TRACENAME="None"):
        """Displays the entire trace in the diagram area, leaving an appropriate display margin.
    
         The trace can be referenced either by its number <WndTr> or by its name <trace_name>.
    
         DISPlay[:WINDow<Wnd>]:TRACe<WndTr>:Y[:SCALe]:AUTO ONCE[, '<trace_name>']
         """
        scpi_command = scpi_preprocess(":DISP:WIND{:}:TRAC{:}:Y:SCAL:AUTO ONCE,'{:}'", wnum, tnum, TRACENAME)
        self.write(scpi_command)

    def set_disp_trace_bottom(self, wnum=1, tnum=1, lower_value="", TRACENAME="None"):
        """Sets the lower (minimum) edge of the diagram area <Wnd>."""
        scpi_command = scpi_preprocess(":DISP:WIND{:}:TRAC{:}:Y:SCAL:BOTT {:},{:}", wnum, tnum, lower_value, TRACENAME)
        self.write(scpi_command)

    def set_disp_trace_delete(self, wnum=1, tnum=1):
        """Releases the assignment between a trace and a diagram area.
    
         As defined by means of DISPlay:WINDow<Wnd>:TRACe<WndTr>:FEED <Trace_Name>
         and expressed by the <WndTr> suffix. The trace itself is not deleted; this
         must be done via CALCulate<Ch>:PARameter:DELete <Trace_Name>.
         """
        scpi_command = scpi_preprocess(":DISP:WIND{:}:TRAC{:}:DEL", wnum, tnum)
        self.write(scpi_command)

    def set_disp_trace_efeed(self, wnum=1, tnum=1, TRACENAME=""):
        """Assigns an existing trace to a diagram area.
    
         Assigns an existing trace (CALCulate<Ch>:PARameter:SDEFine <Trace_Name>)
         to a diagram area <Wnd>, and displays the trace. Use
         DISPlay[:WINDow<Wnd>]:TRACe<WndTr>:FEED to assign the trace to a diagram area
         using a numeric suffix (e.g. in order to use the
         DISPlay[:WINDow<Wnd>]:TRACe<WndTr>:Y:OFFSet command).
    
         DISPlay[:WINDow<Wnd>]:TRACe<WndTr>:EFEed '<trace_name>'
         """
        scpi_command = scpi_preprocess(":DISP:WIND{:}:TRAC{:}:EFE '{:}'", wnum, tnum, TRACENAME)
        self.write(scpi_command)

    def set_disp_trace_feed(self, wnum=1, tnum=1, TRACENAME=""):
        """Assigns an existing trace to a diagram area.
    
         Assigns an existing trace (CALCulate<Ch>:PARameter:SDEFine <Trace_Name>)
         to a diagram area, using the <WndTr> suffix, and displays the trace. Use
         DISPlay[:WINDow<Wnd>]:TRACe:EFEed to assign the trace to a diagram area
         without using a numeric suffix.
         """
        scpi_command = scpi_preprocess(":DISP:WIND{:}:TRAC{:}:FEED '{:}'", wnum, tnum, TRACENAME)
        self.write(scpi_command)

    def set_disp_trace_pdiv(self, wnum=1, tnum=1, value="", TRACENAME="None"):
        """Sets the value between two grid lines (value “per division”) for the diagram area <Wnd>.
    
         DISPlay[:WINDow<Wnd>]:TRACe<WndTr>:Y[:SCALe]:PDIVision  <numeric_value>[, '<trace_name>']
         """
        scpi_command = scpi_preprocess(":DISP:WIND{:}:TRAC{:}:Y:SCAL:PDIV {:},{:}", wnum, tnum, value, TRACENAME)
        self.write(scpi_command)

    def set_disp_trace_reflevel(self, wnum=1, tnum=1, value="", TRACENAME="None"):
        """Sets the reference level (or reference value) for a particular displayed trace.
    
         DISPlay[:WINDow<Wnd>]:TRACe<WndTr>:Y[:SCALe]:RLEVel  <numeric_value>[, '<trace_name>']
         """
        scpi_command = scpi_preprocess(":DISP:WIND{:}:TRAC{:}:Y:SCAL:RLEV {:},{:}", wnum, tnum, value, TRACENAME)
        self.write(scpi_command)

    def set_disp_trace_refpos(self, wnum=1, tnum=1, value="", TRACENAME="None"):
        """Sets the point on the y-axis to be used as the reference position
    
         (as a percentage of the length of the y-axis).
    
         DISPlay[:WINDow<Wnd>]:TRACe<WndTr>:Y[:SCALe]:RPOSition  <numeric_value>[, '<trace_name>']
         """
        scpi_command = scpi_preprocess(":DISP:WIND{:}:TRAC{:}:Y:SCAL:RPOS {:},{:}", wnum, tnum, value, TRACENAME)
        self.write(scpi_command)

    def set_disp_trace_show(self, wnum=1, tnum=1, DMTRACES="", TRACENAME=""):
        """Displays or hides an existing trace, identified by its trace name, or a group of traces.
    
         Displays or hides an existing trace, identified by its trace name
         (CALCulate<Ch>:PARameter:SDEFine <Trace_Name>), or a group of traces.
    
         DISPlay[:WINDow<Wnd>]:TRACe<WndTr>:SHOW DALL | MALL | '<trace_name>', <Boolean>
         """
        scpi_command = scpi_preprocess(":DISP:WIND{:}:TRAC{:}:SHOW {:} {:}", wnum, tnum, DMTRACES, TRACENAME)
        self.write(scpi_command)

    def set_disp_trace_top(self, wnum=1, tnum=1, upper_value="", TRACENAME="None"):
        """Sets the upper (maximum) edge of the diagram area <Wnd>.
    
         DISPlay[:WINDow<Wnd>]:TRACe<WndTr>:Y[:SCALe]:TOP  <upper_value>[, '<trace_name>']
         """
        scpi_command = scpi_preprocess(":DISP:WIND{:}:TRAC{:}:Y:SCAL:TOP {:},{:}", wnum, tnum, upper_value, TRACENAME)
        self.write(scpi_command)

    def set_disp_trace_x_offset(self, wnum=1, tnum=1, range=""):
        """Shifts the trace <WndTr> in horizontal direction, leaving the positions of all markers unchanged."""
        scpi_command = scpi_preprocess(":DISP:WIND{:}:TRAC{:}:X:OFFS {:}", wnum, tnum, range)
        self.write(scpi_command)

    def set_disp_trace_y_offset(self, wnum=1, tnum=1, MAG="", PHASE=0, REAL=0, IMAG=0):
        """Modifies all points of the trace <WndTr> by means of an added and/or a multiplied complex constant."""
        scpi_command = scpi_preprocess(":DISP:WIND{:}:TRAC{:}:Y:OFFS {:}{:}{:}{:}", wnum, tnum, MAG, PHASE, REAL, IMAG)
        self.write(scpi_command)

    def set_disp_update(self, UPDATE="OFF"):
        """no help available"""
        scpi_command = scpi_preprocess(":SYST:DISP:UPD {:}", UPDATE)
        self.write(scpi_command)

    def set_f_start(self, cnum=1, freq=""):
        """Defines the start frequency for a frequency sweep which is equal to the left edge of a Cartesian diagram.
    
         [SENSe<Ch>:]FREQuency:STARt <start_frequency> 
        """
        scpi_command = scpi_preprocess(":SENS{:}:FREQ:STAR {:}", cnum, freq)
        self.write(scpi_command)

    def set_f_stop(self, cnum=1, freq=""):
        """Defines the stop frequency for a frequency sweep which is equal to the right edge of a Cartesian diagram.
    
         [SENSe<Ch>:]FREQuency:STOP <stop_frequency> 
        """
        scpi_command = scpi_preprocess(":SENS{:}:FREQ:STOP {:}", cnum, freq)
        self.write(scpi_command)

    def set_format_binary(self, ORDER="SWAP"):
        """Controls whether binary data is transferred in normal or swapped byte order.
    
         FORMat:BORDer NORMal | SWAPped
         """
        scpi_command = scpi_preprocess(":FORM:BORD {:}", ORDER)
        self.write(scpi_command)

    def set_format_data(self, DATA="ASCII"):
        """Selects the format for numeric data transferred to and from the analyzer.
    
         FORMat[:DATA] ASCii | REAL [,<length>]
    
         NOTE:
         The format setting is only valid for commands and queries whose
         description states that the response is formatted as described by
         FORMat[:DATA]. In particular, it affects trace data transferred by means
         of the commands in the TRACe:... system.
    
         The default length of REAL data is 32 bits (single precision).
         """
        scpi_command = scpi_preprocess(":FORM:DATA {:}", DATA)
        self.write(scpi_command)

    def set_init_continuous(self, cnum=1, STATE="ON"):
        """Qualifies whether the analyzer measures in single sweep or in continuous sweep mode.
    
         INITiate<Ch>:CONTinuous <Boolean>
         """
        scpi_command = scpi_preprocess(":INIT{:}:CONT {:}", cnum, STATE)
        self.write(scpi_command)

    def set_init_immediate(self, cnum=1):
        """Starts a new single sweep sequence.
    
         This command is available in single sweep mode only
         (INITiate<Ch>:CONTinuous OFF). The data of the last sweep (or previous
         sweeps, see Sweep History) can be read using
         CALCulate<Ch>:DATA:NSWeep:FIRSt? SDATa, <count>.
         """
        scpi_command = scpi_preprocess(":INIT{:}:IMM", cnum)
        self.write(scpi_command)

    def set_init_immediate_scope(self, cnum=1, SCOPE="ALL"):
        """Selects the scope of the single sweep sequence.
    
         The setting is applied in single sweep mode only (INITiate<Ch>:CONTinuous OFF).
    
         NITiate<Ch>[:IMMediate]:SCOPe ALL | SINGle
         """
        scpi_command = scpi_preprocess(":INIT{:}:IMM:SCOP {:}", cnum, SCOPE)
        self.write(scpi_command)

    def set_par_del(self, cnum=1, TRACE=""):
        """Deletes a trace with a specified trace name and channel.
    
         CALCulate<Ch>:PARameter:DELete '<trace>'
         """
        scpi_command = scpi_preprocess(":CALC{:}:PAR:DEL '{:}'", cnum, TRACE)
        self.write(scpi_command)

    def set_par_del_all(self, cnum=1):
        """Deletes all traces in all channels of the active setup,
    
         including the default trace Trc1 in channel no. 1. The manual
         control screen shows 'No Trace'
         """
        scpi_command = scpi_preprocess(":CALC{:}:PAR:DEL:ALL", cnum)
        self.write(scpi_command)

    def set_par_del_call(self, cnum=1):
        """Deletes all traces in channel no. <Ch>."""
        scpi_command = scpi_preprocess(":CALC{:}:PAR:DEL:CALL", cnum)
        self.write(scpi_command)

    def set_par_del_sgroup(self, cnum=1):
        """Deletes a group of logical ports (S-parameter group),
    
         previously defined via CALCulate<Ch>:PARameter:DEFine:SGRoup.
         """
        scpi_command = scpi_preprocess(":CALC{:}:PAR:DEL:SGR", cnum)
        self.write(scpi_command)

    def set_par_sdef(self, cnum=1, TRACE="", COEFF=""):
        """Creates a trace and assigns a channel number, a name and a measured quantity to it.
    
         The trace becomes the active trace in the channel but is not displayed.
    
         To select an existing trace as the active trace, use
         CALCulate:PARameter:SELect. You can open the trace manager
         (DISPlay:MENU:KEY:EXECute 'Trace Manager') to obtain an overview of
         all channels and traces, including the traces that are not displayed.
         """
        scpi_command = scpi_preprocess(":CALC{:}:PAR:SDEF '{:}', '{:}'", cnum, TRACE, COEFF)
        self.write(scpi_command)

    def set_par_select(self, cnum=1, TRACE=""):
        """Selects an existing trace as the active trace of the channel.
    
         All trace commands without explicit reference to the trace name act on
         the active trace (e.g. CALCulate<Chn>:FORMat).
         CALCulate<Ch>:PARameter:SELect is also necessary if the active trace of
         a channel has been deleted.
         """
        scpi_command = scpi_preprocess(":CALC{:}:PAR:SEL '{:}'", cnum, TRACE)
        self.write(scpi_command)

    def set_rbw(self, cnum=1, BW=""):
        """Defines the resolution bandwidth of the analyzer (Meas. Bandwidth).
    
         [SENSe<Ch>:]BANDwidth|BWIDth[:RESolution] <bandwidth>
    
         <Ch>  Channel number. If unspecified the numeric suffix is set to 1.
         <bandwidth>
         Resolution bandwidth
        """
        scpi_command = scpi_preprocess(":SENS{:}:BAND:RES {:}", cnum, BW)
        self.write(scpi_command)

    def set_rbw_reduction(self, cnum=1, ONOFF="OFF"):
        """Enables or disables dynamic bandwidth reduction at low frequencies.
    
         [SENSe<Ch>:]BANDwidth|BWIDth[:RESolution]:DREDuction <Boolean>
    
         <Ch>        Channel number
         <Boolean>    ON | OFF – enable or disable dynamic bandwidth reduction.
        """
        scpi_command = scpi_preprocess(":SENS{:}:BAND:DRED {:}", cnum, ONOFF)
        self.write(scpi_command)

    def set_rbw_select(self, cnum=1):
        """Defines the selectivity of the IF filter for an unsegmented sweep.
    
         The value is also used for all segments of a segmented sweep,
         provided that separate selectivity setting is disabled
         ([SENSe<Ch>:]SEGMent<Seg>:BWIDth[:RESolution]:SELect:CONTrol OFF).
    
         [SENSe<Ch>:]BANDwidth|BWIDth[:RESolution]:SELect NORMal | HIGH
        """
        scpi_command = scpi_preprocess(":SENS{:}:BAND:SEL", cnum)
        self.write(scpi_command)

    def set_sweep_count(self, cnum=1, NUMSWEEP=1):
        """Defines the number of sweeps to be measured in single sweep mode
    
         (INITiate<Ch>:CONTinuous OFF).
         """
        scpi_command = scpi_preprocess(":SENS{:}:SWE:COUN {:}", cnum, NUMSWEEP)
        self.write(scpi_command)

    def set_sweep_n_points(self, cnum=1, NUMPOINTS=201):
        """Defines the total number of measurement points per sweep (Number of Points)."""
        scpi_command = scpi_preprocess(":SENS{:}:SWE:POIN {:}", cnum, NUMPOINTS)
        self.write(scpi_command)

    def set_sweep_spacing(self, cnum=1, SPACING="LIN"):
        """Defines the frequency vs. time characteristics of a frequency sweep
    
         (Lin. Frequency or Log Frequency). The command has no effect on segmented
         frequency, power or time sweeps.
    
         [SENSe<Ch>:]SWEep:SPACing  LINear | LOGarithmic
         """
        scpi_command = scpi_preprocess(":SENS{:}:SWE:SPAC {:}", cnum, SPACING)
        self.write(scpi_command)

    def set_sweep_step(self, cnum=1, step_size=""):
        """Sets the distance between two consecutive sweep points.
    
         [SENSe<Ch>:]SWEep:STEP <step_size>
         """
        scpi_command = scpi_preprocess(":SENS{:}:SWE:STEP {:}", cnum, step_size)
        self.write(scpi_command)

    def set_sweep_time(self, cnum=1, duration=""):
        """Sets the duration of the sweep (Sweep Time).
    
         Setting a duration disables the automatic calculation of the (minimum) sweep time;
         see [SENSe<Ch>:]SWEep:TIME:AUTO.
    
         [SENSe<Ch>:]SWEep:TIME <duration>
         """
        scpi_command = scpi_preprocess(":SENS{:}:SWE:TIME {:}", cnum, duration)
        self.write(scpi_command)

    def set_sweep_time_auto(self, cnum=1, ONOFF="ON"):
        """When enabled, the (minimum) sweep duration is calculated internally
    
         using the other channel settings and zero delay ([SENSe<Ch>:]SWEep:DWELl).
    
         [SENSe<Ch>:]SWEep:TIME:AUTO <Boolean>
         """
        scpi_command = scpi_preprocess(":SENS{:}:SWE:TIME:AUTO {:}", cnum, ONOFF)
        self.write(scpi_command)

    def set_sweep_type(self, cnum=1, TYPE="LIN"):
        """Selects the sweep type,
    
         i.e. the sweep variable (frequency/power/time) and the position of the
         sweep points across the sweep range.
    
         [SENSe<Ch>:]SWEep:TYPE LINear | LOGarithmic | SEGMent | POWer | CW | POINt | PULSe | IAMPlitude | IPHase
         """
        scpi_command = scpi_preprocess(":SENS{:}:SWE:TYPE {:}", cnum, TYPE)
        self.write(scpi_command)

    def query_active_channel(self):
        """Selects channel as the active channel.
    
         INSTrument:NSELect <Channel>
         """
        scpi_command = ":INST:NSEL?"
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='int')
        return value

    def query_average_factor_current(self, cnum=1):
        """no help available"""
        scpi_command = scpi_preprocess(":SENS{:}:AVER:COUN:CURR?", cnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='int')
        return value

    def query_averaging_state(self, cnum=1):
        """no help available"""
        scpi_command = scpi_preprocess(":SENS{:}:AVER:STAT?", cnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='bool')
        return value

    def query_channel_catalog(self, cnum=1):
        """Returns the numbers and names of all channels in the current setup."""
        scpi_command = scpi_preprocess(":CONF:CHAN{:}:CAT?", cnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='str')
        return value

    def query_channel_name(self, cnum=1):
        """Assigns a name to channel number <Ch>.
    
         The channel must be created before (CONFigure:CHANnel<Ch>[:STATe] ON).
         Moreover it is not possible to assign the same name to two
         different channels. CONFigure:CHANnel<Ch>:CATalog? returns a list of
         all defined channels with their names.
    
         CONFigure:CHANnel<Ch>:NAME '<Ch_name>'
         """
        scpi_command = scpi_preprocess(":CONF:CHAN{:}:NAME?", cnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='str')
        return value

    def query_channel_name_id(self, cnum=1, CNAME="Ch1"):
        """Queries the channel number (numeric suffix) of a channel with known channel name.
    
         A channel name must be assigned before (CONFigure:CHANnel<Ch>NAME '<Ch_name>').
         CONFigure:CHANnel<Ch>:CATalog? returns a list of all defined
         channels with their names.
    
         CONFigure:CHANnel<Ch>:NAME:ID? '<Ch_name>'
         """
        scpi_command = scpi_preprocess(":CONF:CHAN{:}:NAME:ID? '{:}'", cnum, CNAME)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='str')
        return value

    def query_channel_state(self, cnum=1):
        """Creates or deletes channel no. <Ch> and selects it as the active channel.
    
         CONFigure:CHANnel<Ch>:NAME defines the channel name.
         """
        scpi_command = scpi_preprocess(":CONF:CHAN{:}:STAT?", cnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='str')
        return value

    def query_converter_mode(self, cnum=1):
        """Selects the test setup (internal or external sources)
    
         for the frequency converter measurement (with option ZVA-K8, Converter Control).
    
         [SENSe<Ch>:]FREQuency:CONVersion:DEVice:MODE RILI | RILE | RILI4 | RILI56
    
         The available test setups depend on the instrument type:
    
                             RILI    RILE    RILI4    RILI56
         ---------------------------------------------------
         R&S ZVA 24|40|50      x      x
         R&S ZVA 67                   x        x
         R&S ZVT 20                            x        x
    
         <Ch>    Channel number. This suffix is ignored, the command affects all channels.
         RILI    RF internal, LO internal (4-Port)
                 Ports 1 and 2: Converter RF Signal
                 Ports 3 and 4: Converter LO
         RILE    RF internal, LO external (4-Port)
                 Ports 1 to 4: Converter RF Signal
                 External Generator #1: Converter LO
         RILI4    RF internal, LO internal, Port4=LO (R&S ZVA 67 only)
                 Ports 1 and 2: Converter RF Signal
                 Port 4: Converter LO
         RILI56    RF internal, LO internal, 6-Port (R&S ZVT 20 only)
                 Ports 1 to 4: Converter RF Signal
                 Ports 5 and 6: Converter LO 
        """
        scpi_command = scpi_preprocess(":SENS{:}:FREQ:CONV:DEV:MODE?", cnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='str')
        return value

    def query_converter_name(self, cnum=1):
        """Selects the frequency converter type for enhanced frequency-converting measurements
    
         (with option ZVA-K8, Converter Control).
    
         The preset configuration will be reset to Instrument scope (SYSTem:PRESet:SCOPe ALL)
         and Factory Preset (SYSTem:PRESet:USER:STATe OFF).
    
         [SENSe<Ch>:]FREQuency:CONVersion:DEVice:NAME '<Converter Type>'
    
         | TYPE      | frequency range    |
         |-----------|--------------------|
         | R&S®ZC170 | 110 GHz to 170 GHz |
         | R&S®ZC220 | 140 GHz to 220 GHz |
         | R&S®ZC330 | 220 GHz to 330 GHz |
         | R&S®ZC500 | 330 GHz to 500 GHz |
         |-----------|--------------------|
     
        """
        scpi_command = scpi_preprocess(":SENS{:}:FREQ:CONV:DEV:NAME?", cnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='str')
        return value

    def query_corr_connection(self, cnum=1, pnum=1):
        """no help available"""
        scpi_command = scpi_preprocess(":SENS{:}:CORR:COLL:CONN{:}?", cnum, pnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='str')
        return value

    def query_corr_connection_genders(self, cnum=1, pnum=1):
        """no help available"""
        scpi_command = scpi_preprocess(":SENS{:}:CORR:COLL:CONN{:}:GEND?", cnum, pnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='str')
        return value

    def query_corr_connection_ports(self, cnum=1, pnum=1):
        """no help available"""
        scpi_command = scpi_preprocess(":SENS{:}:CORR:COLL:CONN{:}:PORT?", cnum, pnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='str')
        return value

    def query_corr_date(self, cnum=1):
        """no help available"""
        scpi_command = scpi_preprocess(":SENS{:}:CORR:DATE?", cnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='str')
        return value

    def query_corr_state(self, cnum=1):
        """no help available"""
        scpi_command = scpi_preprocess(":SENS{:}:CORR:STAT?", cnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='str')
        return value

    def query_data(self, cnum=1, SDATA=""):
        """no help available"""
        scpi_command = scpi_preprocess(":CALC{:}:DATA? {:}", cnum, SDATA)
        return self.query_values(scpi_command)

    def query_data_all(self, cnum=1, SDATA=""):
        """no help available"""
        scpi_command = scpi_preprocess(":CALC{:}:DATA:ALL? {:}", cnum, SDATA)
        return self.query_values(scpi_command)

    def query_data_call(self, cnum=1, SDATA=""):
        """no help available"""
        scpi_command = scpi_preprocess(":CALC{:}:DATA:CALL? {:}", cnum, SDATA)
        return self.query_values(scpi_command)

    def query_data_call_catalog(self, cnum=1):
        """no help available"""
        scpi_command = scpi_preprocess(":CALC{:}:DATA:CALL:CAT?", cnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=True, strip_outer_quotes=True, returns='str')
        return value

    def query_data_dall(self, cnum=1, SDATA=""):
        """no help available"""
        scpi_command = scpi_preprocess(":CALC{:}:DATA:DALL? {:}", cnum, SDATA)
        return self.query_values(scpi_command)

    def query_data_sgroup(self, cnum=1, SDATA=""):
        """no help available"""
        scpi_command = scpi_preprocess(":CALC{:}:DATA:SGR? {:}", cnum, SDATA)
        return self.query_values(scpi_command)

    def query_disp_catalog(self, wnum=1):
        """Returns the numbers and names of all diagram areas in the current setup."""
        scpi_command = scpi_preprocess(":DISP:WIND{:}:CAT?", wnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='str')
        return value

    def query_disp_name(self, wnum=1):
        """Defines a name for diagram area <Wnd>.
    
         The name appears in the list of diagram areas, to be queried by DISPlay[:WINDow<Wnd>]:CAT?.
         """
        scpi_command = scpi_preprocess(":DISP:WIND{:}:NAME?", wnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='str')
        return value

    def query_disp_state(self, wnum=1):
        """Creates or deletes a diagram area, identified by its area number <Wnd>.
    
         DISPlay[:WINDow<Wnd>]:STATe <Boolean>
         """
        scpi_command = scpi_preprocess(":DISP:WIND{:}:STAT?", wnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='bool')
        return value

    def query_disp_title_data(self, wnum=1):
        """Defines a title for diagram area <Wnd>.
    
         DISPlay[:WINDow<Wnd>]:TITLe:DATA '<string>'
         """
        scpi_command = scpi_preprocess(":DISP:WIND{:}:TITL:DATA?", wnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='str')
        return value

    def query_disp_title_state(self, wnum=1):
        """Displays or hides the title for area number <Wnd>, defined by means of DISPlay:WINDow<Wnd>:TITLe:DATA.
    
         DISPlay[:WINDow<Wnd>]:TITLe[:STATe] <Boolean>
         """
        scpi_command = scpi_preprocess(":DISP:WIND{:}:TITL:STAT?", wnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='bool')
        return value

    def query_disp_trace_catalog(self, wnum=1, tnum=1):
        """Returns the numbers and names of all traces in diagram area no. <Wnd>."""
        scpi_command = scpi_preprocess(":DISP:WIND{:}:TRAC{:}:CAT?", wnum, tnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='str')
        return value

    def query_disp_trace_show(self, wnum=1, tnum=1, DMTRACES=""):
        """Displays or hides an existing trace, identified by its trace name, or a group of traces.
    
         Displays or hides an existing trace, identified by its trace name
         (CALCulate<Ch>:PARameter:SDEFine <Trace_Name>), or a group of traces.
    
         DISPlay[:WINDow<Wnd>]:TRACe<WndTr>:SHOW DALL | MALL | '<trace_name>', <Boolean>
         """
        scpi_command = scpi_preprocess(":DISP:WIND{:}:TRAC{:}:SHOW? {:}", wnum, tnum, DMTRACES)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='bool')
        return value

    def query_disp_trace_x_offset(self, wnum=1, tnum=1):
        """Shifts the trace <WndTr> in horizontal direction, leaving the positions of all markers unchanged."""
        scpi_command = scpi_preprocess(":DISP:WIND{:}:TRAC{:}:X:OFFS?", wnum, tnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='str')
        return value

    def query_disp_trace_y_offset(self, wnum=1, tnum=1):
        """Modifies all points of the trace <WndTr> by means of an added and/or a multiplied complex constant."""
        scpi_command = scpi_preprocess(":DISP:WIND{:}:TRAC{:}:Y:OFFS?", wnum, tnum)
        return self.query_values(scpi_command)

    def query_f_start(self, cnum=1):
        """Defines the start frequency for a frequency sweep which is equal to the left edge of a Cartesian diagram.
    
         [SENSe<Ch>:]FREQuency:STARt <start_frequency> 
        """
        scpi_command = scpi_preprocess(":SENS{:}:FREQ:STAR?", cnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='float')
        return value

    def query_f_stop(self, cnum=1):
        """Defines the stop frequency for a frequency sweep which is equal to the right edge of a Cartesian diagram.
    
         [SENSe<Ch>:]FREQuency:STOP <stop_frequency> 
        """
        scpi_command = scpi_preprocess(":SENS{:}:FREQ:STOP?", cnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='float')
        return value

    def query_format_binary(self):
        """Controls whether binary data is transferred in normal or swapped byte order.
    
         FORMat:BORDer NORMal | SWAPped
         """
        scpi_command = ":FORM:BORD?"
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='str')
        return value

    def query_format_data(self):
        """Selects the format for numeric data transferred to and from the analyzer.
    
         FORMat[:DATA] ASCii | REAL [,<length>]
    
         NOTE:
         The format setting is only valid for commands and queries whose
         description states that the response is formatted as described by
         FORMat[:DATA]. In particular, it affects trace data transferred by means
         of the commands in the TRACe:... system.
    
         The default length of REAL data is 32 bits (single precision).
         """
        scpi_command = ":FORM:DATA?"
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='str')
        return value

    def query_init_continuous(self, cnum=1):
        """Qualifies whether the analyzer measures in single sweep or in continuous sweep mode.
    
         INITiate<Ch>:CONTinuous <Boolean>
         """
        scpi_command = scpi_preprocess(":INIT{:}:CONT?", cnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='bool')
        return value

    def query_init_immediate_scope(self, cnum=1):
        """Selects the scope of the single sweep sequence.
    
         The setting is applied in single sweep mode only (INITiate<Ch>:CONTinuous OFF).
    
         NITiate<Ch>[:IMMediate]:SCOPe ALL | SINGle
         """
        scpi_command = scpi_preprocess(":INIT{:}:IMM:SCOP?", cnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='str')
        return value

    def query_par_catalog(self, cnum=1):
        """Returns the trace names and measured quantities of all traces assigned to a particular channel."""
        scpi_command = scpi_preprocess(":CALC{:}:PAR:CAT?", cnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=True, strip_outer_quotes=True, returns='str')
        return value

    def query_par_select(self, cnum=1):
        """Selects an existing trace as the active trace of the channel.
    
         All trace commands without explicit reference to the trace name act on
         the active trace (e.g. CALCulate<Chn>:FORMat).
         CALCulate<Ch>:PARameter:SELect is also necessary if the active trace of
         a channel has been deleted.
         """
        scpi_command = scpi_preprocess(":CALC{:}:PAR:SEL?", cnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='str')
        return value

    def query_rbw(self, cnum=1):
        """Defines the resolution bandwidth of the analyzer (Meas. Bandwidth).
    
         [SENSe<Ch>:]BANDwidth|BWIDth[:RESolution] <bandwidth>
    
         <Ch>  Channel number. If unspecified the numeric suffix is set to 1.
         <bandwidth>
         Resolution bandwidth
        """
        scpi_command = scpi_preprocess(":SENS{:}:BAND:RES?", cnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='str')
        return value

    def query_rbw_reduction(self, cnum=1):
        """Enables or disables dynamic bandwidth reduction at low frequencies.
    
         [SENSe<Ch>:]BANDwidth|BWIDth[:RESolution]:DREDuction <Boolean>
    
         <Ch>        Channel number
         <Boolean>    ON | OFF – enable or disable dynamic bandwidth reduction.
        """
        scpi_command = scpi_preprocess(":SENS{:}:BAND:DRED?", cnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='bool')
        return value

    def query_rbw_select(self, cnum=1):
        """Defines the selectivity of the IF filter for an unsegmented sweep.
    
         The value is also used for all segments of a segmented sweep,
         provided that separate selectivity setting is disabled
         ([SENSe<Ch>:]SEGMent<Seg>:BWIDth[:RESolution]:SELect:CONTrol OFF).
    
         [SENSe<Ch>:]BANDwidth|BWIDth[:RESolution]:SELect NORMal | HIGH
        """
        scpi_command = scpi_preprocess(":SENS{:}:BAND:SEL?", cnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='str')
        return value

    def query_sweep_count(self, cnum=1):
        """Defines the number of sweeps to be measured in single sweep mode
    
         (INITiate<Ch>:CONTinuous OFF).
         """
        scpi_command = scpi_preprocess(":SENS{:}:SWE:COUN?", cnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='int')
        return value

    def query_sweep_n_points(self, cnum=1):
        """Defines the total number of measurement points per sweep (Number of Points)."""
        scpi_command = scpi_preprocess(":SENS{:}:SWE:POIN?", cnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='int')
        return value

    def query_sweep_spacing(self, cnum=1):
        """Defines the frequency vs. time characteristics of a frequency sweep
    
         (Lin. Frequency or Log Frequency). The command has no effect on segmented
         frequency, power or time sweeps.
    
         [SENSe<Ch>:]SWEep:SPACing  LINear | LOGarithmic
         """
        scpi_command = scpi_preprocess(":SENS{:}:SWE:SPAC?", cnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='str')
        return value

    def query_sweep_step(self, cnum=1):
        """Sets the distance between two consecutive sweep points.
    
         [SENSe<Ch>:]SWEep:STEP <step_size>
         """
        scpi_command = scpi_preprocess(":SENS{:}:SWE:STEP?", cnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='str')
        return value

    def query_sweep_time(self, cnum=1):
        """Sets the duration of the sweep (Sweep Time).
    
         Setting a duration disables the automatic calculation of the (minimum) sweep time;
         see [SENSe<Ch>:]SWEep:TIME:AUTO.
    
         [SENSe<Ch>:]SWEep:TIME <duration>
         """
        scpi_command = scpi_preprocess(":SENS{:}:SWE:TIME?", cnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='str')
        return value

    def query_sweep_time_auto(self, cnum=1):
        """When enabled, the (minimum) sweep duration is calculated internally
    
         using the other channel settings and zero delay ([SENSe<Ch>:]SWEep:DWELl).
    
         [SENSe<Ch>:]SWEep:TIME:AUTO <Boolean>
         """
        scpi_command = scpi_preprocess(":SENS{:}:SWE:TIME:AUTO?", cnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='bool')
        return value

    def query_sweep_type(self, cnum=1):
        """Selects the sweep type,
    
         i.e. the sweep variable (frequency/power/time) and the position of the
         sweep points across the sweep range.
    
         [SENSe<Ch>:]SWEep:TYPE LINear | LOGarithmic | SEGMent | POWer | CW | POINt | PULSe | IAMPlitude | IPHase
         """
        scpi_command = scpi_preprocess(":SENS{:}:SWE:TYPE?", cnum)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='str')
        return value

    def query_system_err_next(self):
        """no help available"""
        scpi_command = ":SYST:ERR:NEXT?"
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='str')
        return value

    def query_system_error_all(self):
        """no help available"""
        scpi_command = ":SYST:ERR:ALL?"
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='str')
        return value



import re

null_parameter = re.compile(",{2,}")  # detect optional null parameter as two consecutive commas, and remove
converters = {
    "str": str,
    "int": int,
    "float": float,
    "bool": lambda x: bool(int(x)),
}


def to_string(value):
    tval = type(value)
    if tval is str:
        return value
    elif tval is bool:
        return str(int(value))
    elif tval in (list, tuple):
        return ",".join(map(to_string, value))
    elif value is None:
        return ""
    else:
        return str(value)


def scpi_preprocess(command_string, *args):
    args = list(args)
    for i, arg in enumerate(args):
        args[i] = to_string(arg)
    cmd = command_string.format(*args)
    return null_parameter.sub(",", cmd)


def process_query(query, csv=False, strip_outer_quotes=True, returns="str"):
    if strip_outer_quotes is True:
        if query[0] + query[-1] in ('""', "''"):
            query = query[1:-1]
    if csv is True:
        query = query.split(",")

    converter = None if returns == "str" else converters.get(returns, None)
    if converter:
        if csv is True:
            query = list(map(converter, query))
        else:
            query = converter(query)

    return query


class SCPI(object):
    def __init__(self, resource):
        self.resource = resource
        self.echo = False  # print scpi command string to scpi out

    def write(self, scpi, *args, **kwargs):
        if self.echo:
            print(scpi)
        self.resource.write(scpi, *args, **kwargs)

    def query(self, scpi, *args, **kwargs):
        if self.echo:
            print(scpi)
        return self.resource.query(scpi, *args, **kwargs)

    def write_values(self, scpi, *args, **kwargs):
        if self.echo:
            print(scpi)
        self.resource.write_values(scpi, *args, **kwargs)

    def query_values(self, scpi, *args, **kwargs):
        if self.echo:
            print(scpi)
        return self.resource.query_values(scpi, *args, **kwargs)

    def set_averaging_count(self, avg_count=""):
        """no help available"""
        scpi_command = scpi_preprocess(":AVER:COUN {:}", avg_count)
        self.write(scpi_command)

    def set_averaging_mode(self, avg_mode="SWEEP"):
        """no help available"""
        scpi_command = scpi_preprocess(":AVER:MODE {:}", avg_mode)
        self.write(scpi_command)

    def set_clear_averaging(self):
        """no help available"""
        scpi_command = ":AVER:CLE"
        self.write(scpi_command)

    def set_continuous_sweep(self, onoff="ON"):
        """no help available"""
        scpi_command = scpi_preprocess(":INIT:CONT {:}", onoff)
        self.write(scpi_command)

    def set_current_trace(self, trace=1):
        """no help available"""
        scpi_command = scpi_preprocess(":CALC:PAR{:}:SEL", trace)
        self.write(scpi_command)

    def set_display_format(self, fmt="MLOG"):
        """no help available"""
        scpi_command = scpi_preprocess(":CALC:FORM {:}", fmt)
        self.write(scpi_command)

    def set_f_start(self, freq=""):
        """no help available"""
        scpi_command = scpi_preprocess(":FREQ:START {:}", freq)
        self.write(scpi_command)

    def set_f_stop(self, freq=""):
        """no help available"""
        scpi_command = scpi_preprocess(":FREQ:STOP {:}", freq)
        self.write(scpi_command)

    def set_if_bandwidth(self, bandwidth=1000):
        """no help available"""
        scpi_command = scpi_preprocess(":BWID {:}", bandwidth)
        self.write(scpi_command)

    def set_instrument(self, instr="NA"):
        """no help available"""
        scpi_command = scpi_preprocess(":INST '{:}'", instr)
        self.write(scpi_command)

    def set_single_sweep(self):
        """no help available"""
        scpi_command = ":INIT"
        self.write(scpi_command)

    def set_sweep_n_points(self, n_points=401):
        """no help available"""
        scpi_command = scpi_preprocess(":SWE:POIN {:}", n_points)
        self.write(scpi_command)

    def set_trace_autoscale(self, tnum=1):
        """no help available"""
        scpi_command = scpi_preprocess(":DISP:WIND:TRAC{:}:Y:AUTO", tnum)
        self.write(scpi_command)

    def set_trace_count(self, num=4):
        """no help available"""
        scpi_command = scpi_preprocess(":CALC:PAR:COUN {:}", num)
        self.write(scpi_command)

    def set_trace_display_config(self, config="D12_34"):
        """no help available"""
        scpi_command = scpi_preprocess(":DISP:WIND:SPL {:}", config)
        self.write(scpi_command)

    def set_trace_measurement(self, trace=1, meas='S11'):
        """no help available"""
        scpi_command = scpi_preprocess(":CALC:PAR{:}:DEF {:}", trace, meas)
        self.write(scpi_command)

    def query_averaging_count(self):
        """no help available"""
        scpi_command = ":AVER:COUN?"
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='int')
        return value

    def query_averaging_mode(self):
        """no help available"""
        scpi_command = ":AVER:MODE?"
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='str')
        return value

    def query_continuous_sweep(self):
        """no help available"""
        scpi_command = ":INIT:CONT?"
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='bool')
        return value

    def query_current_trace_data(self):
        """no help available"""
        scpi_command = ":CALC:DATA:SDAT?"
        return self.query_values(scpi_command)

    def query_display_format(self):
        """no help available"""
        scpi_command = ":CALC:FORM?"
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='str')
        return value

    def query_f_start(self):
        """no help available"""
        scpi_command = ":FREQ:START?"
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='float')
        return value

    def query_f_stop(self):
        """no help available"""
        scpi_command = ":FREQ:STOP?"
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='float')
        return value

    def query_if_bandwidth(self):
        """no help available"""
        scpi_command = ":BWID?"
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='int')
        return value

    def query_instrument(self):
        """no help available"""
        scpi_command = ":INST?"
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='str')
        return value

    def query_sweep_n_points(self):
        """no help available"""
        scpi_command = ":SWE:POIN?"
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='int')
        return value

    def query_sweep_time(self):
        """no help available"""
        scpi_command = ":SWE:TIME?"
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='float')
        return value

    def query_trace_count(self):
        """no help available"""
        scpi_command = ":CALC:PAR:COUN?"
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='int')
        return value

    def query_trace_display_config(self):
        """no help available"""
        scpi_command = ":DISP:WIND:SPL?"
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='str')
        return value

    def query_trace_measurement(self, trace=1):
        """no help available"""
        scpi_command = scpi_preprocess(":CALC:PAR{:}:DEF?", trace)
        value = self.query(scpi_command)
        value = process_query(value, csv=False, strip_outer_quotes=True, returns='str')
        return value



import unittest

import skrf as rf
import numpy as npy
from skrf.calibration.calibrationSet import Dot
from skrf.calibration import OnePort, TRL, SOLT, EightTerm
from skrf.util import suppress_warning_decorator


class CalsetTest(object):
    @suppress_warning_decorator("No switch terms")
    def test_run(self):
        '''
        ensure cal_set can be generated
        '''
        self.calset.run()

    @suppress_warning_decorator("No switch terms")
    def test_correct_ntwk(self):
        '''
        ensure a network can be corrected
        '''
        ntwk = self.wg.random(n_ports = self.n_ports)
        self.calset.apply_cal(ntwk)

class DotOneport(unittest.TestCase,CalsetTest):
    '''
    

    '''
    def setUp(self):
        self.wg = rf.RectangularWaveguide(rf.F(75,100,11), a=100*rf.mil,z0=50)
        wg = self.wg
        self.n_ports = 1
        self.E = wg.random(n_ports =2, name = 'E')
        
        
        ideals = [
                wg.short( name='short'),
                wg.delay_short( 45.,'deg',name='ew'),
                wg.delay_short( 90.,'deg',name='qw'),
                wg.match( name='load'),
                ]
        measured_sets = []
        for ideal in ideals:
            ns = rf.NetworkSet([self.measure(ideal) for k in range(3)])
            measured_sets.append(ns)
        
        
        self.calset = Dot(cal_class = OnePort, 
                          ideals = ideals, 
                          measured_sets = measured_sets,
                          is_reciprocal = True)
            
    def measure(self, ntwk):
        out = self.E**ntwk
        out.name = ntwk.name
        return out

class DotEightTerm(unittest.TestCase, CalsetTest):
    @suppress_warning_decorator("invalid value encountered in multiply")
    @suppress_warning_decorator("divide by zero encountered in true_divide")
    def setUp(self):
        self.n_ports = 2
        self.wg = rf.RectangularWaveguide(rf.F(75,100,3), a=100*rf.mil,z0=50)
        wg= self.wg
        wg.frequency = rf.F.from_f([100])
        
        self.X = wg.random(n_ports =2, name = 'X')
        self.Y = wg.random(n_ports =2, name='Y')
        self.gamma_f = wg.random(n_ports =1, name='gamma_f')
        self.gamma_r = wg.random(n_ports =1, name='gamma_r')
        
        
        ideals = [
            wg.short(nports=2, name='short'),
            wg.open(nports=2, name='open'),
            wg.match(nports=2, name='load'),
            wg.thru(name='thru'),
            ]
            
        measured_sets = []
        for ideal in ideals:
            ns = rf.NetworkSet([self.measure(ideal) for k in range(3)])
            measured_sets.append(ns)
        
        self.calset = Dot(  cal_class = EightTerm,
                            ideals = ideals,
                            measured_sets = measured_sets,
                            switch_terms = (self.gamma_f, self.gamma_r)
                            )
        
        
 
        
    def measure(self,ntwk):
        out =  self.X**ntwk**self.Y
        out.name = ntwk.name
        return out
    





'''

.. currentmodule:: skrf.instances
========================================
instances (:mod:`skrf.instances`)
========================================

This module contains commonly used  instantiated objects's.



Standard Waveguide Bands
----------------------------

:class:`~skrf.frequency.Frequency` Objects
++++++++++++++++++++++++++++++++++++++++++++
These are predefined :class:`~skrf.frequency.Frequency` objects
that correspond to standard waveguide bands. This information is taken
from the VDI Application Note 1002 [#]_ . IEEE designators are taken
from Spinner TD-00036 [*]_ .


=======================  ===============================================
Object Name              Description
=======================  ===============================================
f_wr10                   WR-10, 75-110 GHz
f_wr3                    WR-3, 220-325 GHz
f_wr2p2                  WR-2.2, 330-500 GHz
f_wr1p5                  WR-1.5, 500-750 GHz
f_wr1                    WR-1, 750-1100 GHz
...                      ...
=======================  ===============================================


:class:`~skrf.media.rectangularWaveguide.RectangularWaveguide`  Objects
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
These are predefined :class:`~skrf.media.rectangularWaveguide.RectangularWaveguide`
objects for  standard waveguide bands.

=======================  ===============================================
Object Name              Description
=======================  ===============================================
wr10                     WR-10, 75-110 GHz
wr3                      WR-3, 220-325 GHz
wr2p2                    WR-2.2, 330-500 GHz
wr1p5                    WR-1.5, 500-750 GHz
wr1                      WR-1, 750-1100 GHz
...                      ...
=======================  ===============================================



References
-------------
.. [#] VDI Application Note:  VDI Waveguide Band Designations (VDI-1002) http://vadiodes.com/VDI/pdf/waveguidechart200908.pdf
.. [*] Spinner Technical Information: Cross Reference For Hollow Metallic Waveguides (TD-00036) https://www.spinner-group.com/images/download/technical_documents/SPINNER_TD00036.pdf
'''

from . frequency import Frequency
from . media import RectangularWaveguide, Freespace, DefinedGammaZ0
from .constants import mil




air = Freespace()
air50 = Freespace(z0=50)


######## waveguide bands
f_wr51  = Frequency(15,22,1001, 'ghz')
f_wr42  = Frequency(17.5,26.5,1001, 'ghz')
f_wr34  = Frequency(22,33,1001, 'ghz')
f_wr28  = Frequency(26.5,40,1001, 'ghz')
f_wr22p4  = Frequency(33,50.5,1001, 'ghz')
f_wr18p8  = Frequency(40,60,1001, 'ghz')
f_wr14p8  = Frequency(50,75,1001, 'ghz')
f_wr12p2  = Frequency(60,90,1001, 'ghz')
f_wr10  = Frequency(75,110,1001, 'ghz')
f_wr8  = Frequency(90,140,1001, 'ghz')
f_wr6p5  = Frequency(110,170,1001, 'ghz')
f_wr5p1  = Frequency(140,220,1001, 'ghz')
f_wr4p3  = Frequency(170,260,1001, 'ghz')
f_wr3p4  = Frequency(220,330,1001, 'ghz')
f_wr2p8 = Frequency(260,400,1001, 'ghz')
f_wr2p2 = Frequency(330,500,1001, 'ghz')
f_wr1p9 = Frequency(400,600,1001, 'ghz')
f_wr1p5 = Frequency(500,750,1001, 'ghz')
f_wr1p2   = Frequency(600,900,1001, 'ghz')
f_wr1   = Frequency(750,1100,1001, 'ghz')
f_wr0p8   = Frequency(900,1400,1001, 'ghz')
f_wr0p65  = Frequency(1100,1700,1001, 'ghz')
f_wr0p51   = Frequency(1400,2200,1001, 'ghz')

f_wm1295 = f_wr5p1
f_wm1092 = f_wr4p3
f_wm864  = f_wr3p4
f_wm710  = f_wr2p8
f_wm570  = f_wr2p2
f_wm470  = f_wr1p9
f_wm380  = f_wr1p5
f_wm310  = f_wr1p2
f_wm250  = f_wr1
f_wm200  = f_wr0p8
f_wm164  = f_wr0p65
f_wm130  = f_wr0p51
f_wm106  = Frequency(1700,2600,1001, 'ghz')
f_wm86   = Frequency(2200,3300,1001, 'ghz')


wr51  = RectangularWaveguide(f_wr51.copy(),a=510*mil,b=255*mil,z0=50)
wr42  = RectangularWaveguide(f_wr42.copy(),a=420*mil,b=170*mil,z0=50)
wr34  = RectangularWaveguide(f_wr34.copy(),a=340*mil,b=170*mil,z0=50)
wr28  = RectangularWaveguide(f_wr28.copy(),a=280*mil,b=140*mil,z0=50)
wr22p4  = RectangularWaveguide(f_wr22p4.copy(),a=224*mil,b=112*mil,z0=50)
wr18p8  = RectangularWaveguide(f_wr18p8.copy(),a=188*mil,b=94*mil,z0=50)
wr14p8  = RectangularWaveguide(f_wr14p8.copy(),a=148*mil,b=74*mil,z0=50)
wr12p2  = RectangularWaveguide(f_wr12p2.copy(),a=122*mil,b=61*mil,z0=50)
wr10  = RectangularWaveguide(f_wr10.copy(),a=100*mil,b=50*mil,z0=50)
wr8  = RectangularWaveguide(f_wr8.copy(),a=80*mil,b=40*mil,z0=50)
wr6p5  = RectangularWaveguide(f_wr6p5.copy(),a=65*mil,b=32.5*mil,z0=50)
wr5p1  = RectangularWaveguide(f_wr5p1.copy(),a=51*mil,b=25.5*mil,z0=50)
wr4p3  = RectangularWaveguide(f_wr4p3.copy(),a=43*mil,b=21.5*mil,z0=50)
wr3p4  = RectangularWaveguide(f_wr3p4.copy(),a=34*mil,b=17*mil,z0=50)
wr2p8 = RectangularWaveguide(f_wr2p8.copy(),a=28*mil,b=14*mil,z0=50)
wr2p2 = RectangularWaveguide(f_wr2p2.copy(),a=22*mil,b=11*mil,z0=50)
wr1p9 = RectangularWaveguide(f_wr1p9.copy(),a=19*mil,b=9.5*mil,z0=50)
wr1p5 = RectangularWaveguide(f_wr1p5.copy(),a=15*mil,b=7.5*mil,z0=50)
wr1p2   = RectangularWaveguide(f_wr1p2.copy(),a=12*mil,b=6*mil,z0=50)
wr1   = RectangularWaveguide(f_wr1.copy(),a=10*mil,b=5*mil,z0=50)
wr0p8   = RectangularWaveguide(f_wr0p8.copy(),a=8*mil,b=4*mil,z0=50)
wr0p65  = RectangularWaveguide(f_wr0p65.copy(),a=6.5*mil,b=3.25*mil,z0=50)
wr0p51   = RectangularWaveguide(f_wr0p51.copy(),a=5.1*mil,b=2.55*mil,z0=50)

wm1295 = RectangularWaveguide(f_wm1295.copy(),a=1295e-6,b=647.5e-6,z0=50)
wm1092 = RectangularWaveguide(f_wm1092.copy(),a=1092e-6,b=546e-6,z0=50)
wm864  = RectangularWaveguide(f_wm864.copy(),a=864e-6,b=432e-6,z0=50)
wm710  = RectangularWaveguide(f_wm710.copy(),a=710e-6,b=355e-6,z0=50)
wm570  = RectangularWaveguide(f_wm570.copy(),a=570e-6,b=285e-6,z0=50)
wm470  = RectangularWaveguide(f_wm470.copy(),a=470e-6,b=235e-6,z0=50)
wm380  = RectangularWaveguide(f_wm380.copy(),a=380e-6,b=190e-6,z0=50)
wm310  = RectangularWaveguide(f_wm310.copy(),a=310e-6,b=155e-6,z0=50)
wm250  = RectangularWaveguide(f_wm250.copy(),a=250e-6,b=125e-6,z0=50)
wm200  = RectangularWaveguide(f_wm200.copy(),a=200e-6,b=100e-6,z0=50)
wm164  = RectangularWaveguide(f_wm164.copy(),a=164e-6,b=82e-6,z0=50)
wm130  = RectangularWaveguide(f_wm130.copy(),a=130e-6,b=65e-6,z0=50)
wm106  = RectangularWaveguide(f_wm106.copy(),a=106e-6,b=53e-6,z0=50)
wm86   = RectangularWaveguide(f_wm86.copy(),a=86e-6,b=43e-6,z0=50)



import warnings
from collections import OrderedDict, Iterable

import numpy as np
import skrf
import pyvisa

from . import abcvna
from . import rs_zva_scpi


class ZVA(abcvna.VNA):
    """
    Class for modern Rohde&Schwarz ZVA Vector Network Analyzers
    """

    NAME = "R&S ZVA"
    NPORTS = 4
    NCHANNELS = 32
    SCPI_VERSION_TESTED = 'unconfirmed'

    def __init__(self, address, **kwargs):
        """
        initialization of ZVA Class

        Parameters
        ----------
        address : str
            visa resource string (full string or ip address)
        kwargs : dict
            interface (str), port (int), timeout (int),
        """
        super(ZVA, self).__init__(address, **kwargs)
        self.resource.timeout = kwargs.get("timeout", 2000)
        self.scpi = rs_zva_scpi.SCPI(self.resource)
        self.use_ascii()  # less likely to cause problems than use_binary
        print(self.idn)

    def use_binary(self):
        """setup the analyzer to transfer in binary which is faster, especially
        for large datasets"""
        self.scpi.set_format_binary(ORDER='SWAP')
        self.scpi.set_format_data(DATA='REAL,64')
        self.resource.values_format.use_binary(datatype='d',
                                               is_big_endian=False,
                                               container=np.array)

    def use_ascii(self):
        self.scpi.set_format_data(DATA='ASCII')
        self.resource.values_format.use_ascii(converter='f', separator=',',
                                              container=np.array)

    @property
    def echo(self):
        return self.scpi.echo

    @echo.setter
    def echo(self, onoff):
        if onoff in (1, True):
            self.scpi.echo = True
        elif onoff in (0, False):
            self.scpi.echo = False
        else:
            raise warnings.warn("echo must be a boolean")

    @property
    def active_channel(self):
        old_timeout = self.resource.timeout
        self.resource.timeout = 500
        try:
            channel = self.scpi.query_active_channel()
        except pyvisa.VisaIOError:
            print("No channel active, using 1")
            channel = 1
        finally:
            self.resource.timeout = old_timeout
        return channel

    @active_channel.setter
    def active_channel(self, channel):
        """
        Set the active channel on the analyzer

        Parameters
        ----------
        channel : int

        Notes
        -----
        """
        old_timeout = self.resource.timeout
        self.resource.timeout = 500
        if channel in self.channel_list:
            self.scpi.set_active_channel(channel)
        else:
            print('Channel %i not in list of channels. Create channel first'
                  % channel)
        return self.scpi.query_active_channel()

    @property
    def channel_list(self):
        """Return list of channels"""
        return_str = self.scpi.query_channel_catalog().split(',')
        channel_dct = {}
        for i in range(int(len(return_str)/2)):
            channel_dct[int(return_str[2 * i])] = return_str[2 * i + 1]
        return channel_dct

    def get_frequency(self, **kwargs):
        """
        get an skrf.Frequency object for the current channel

        Parameters
        ----------
        kwargs : dict
            channel (int), f_unit (str)

        Returns
        -------
        skrf.Frequency
        """
        #self.resource.clear()
        channel = kwargs.get("channel", self.active_channel)
        use_log = "LOG" in self.scpi.query_sweep_type(channel).upper()
        f_start = self.scpi.query_f_start(channel)
        f_stop = self.scpi.query_f_stop(channel)
        f_npoints = self.scpi.query_sweep_n_points(channel)
        if use_log:
            freq = np.logspace(np.log10(f_start), np.log10(f_stop), f_npoints)
        else:
            freq = np.linspace(f_start, f_stop, f_npoints)

        frequency = skrf.Frequency.from_f(freq, unit="Hz")
        frequency.unit = kwargs.get("f_unit", "Hz")
        return frequency

    def set_frequency_sweep(self, f_start, f_stop, f_npoints, **kwargs):
        f_unit = kwargs.get("f_unit", "hz").lower()
        if f_unit != "hz":
            f_start = self.to_hz(f_start, f_unit)
            f_stop = self.to_hz(f_stop, f_unit)
        channel = kwargs.get("channel", self.active_channel)
        self.scpi.set_f_start(channel, f_start)
        self.scpi.set_f_stop(channel, f_stop)
        self.scpi.set_sweep_n_points(channel, f_npoints)

    def get_active_trace_as_network(self, **kwargs):
        """get the current trace as a 1-port network object"""
        channel = self.active_channel
        f_unit = kwargs.get("f_unit", "GHz")
        ntwk = skrf.Network()
        ntwk.name = kwargs.get("name", self.scpi.query_par_select(channel))
        ntwk.frequency = self.get_frequency(channel=channel, f_unit=f_unit)
        sdata = self.scpi.query_data(channel, "SDATA")
        ntwk.s = sdata[::2] + 1j * sdata[1::2]
        return ntwk

    def get_snp_network(self, ports, **kwargs):
        if "sweep" in kwargs.keys():
            warnings.warn("Sweep function not yet implemented for ZVA")

        # ensure all ports are ints, unique and in a valid range
        for i, port in enumerate(ports):
            if type(port) is not int:
                raise TypeError("ports must be an iterable of integers, not type: {:d}".format(type(port)))
            if not 0 < port <= self.NPORTS:
                raise ValueError("invalid ports, must be between 1 and {:d}".format(self.NPORTS))
            if port in ports[i+1:]:
                raise ValueError("duplicate port: {:d}".format(port))

        channel = kwargs.get("channel", self.active_channel)
        catalogue = self.scpi.query_par_catalog(channel)  # type: list
        trace_name = catalogue[::2]
        trace_data = catalogue[1::2]

        port_keys = list()
        for receive_port in ports:
            for source_port in ports:
                key = "S{:d}{:d}".format(receive_port, source_port)
                if key not in trace_data:
                    raise Exception("missing measurement trace for {:s}".format(key))
                port_keys.append(key)

        nports = len(ports)
        npoints = self.scpi.query_sweep_n_points(channel)
        ntwk = skrf.Network()
        f_unit = kwargs.get("f_unit", "GHz")
        ntwk.frequency = self.get_frequency(channel=channel, f_unit="GHz")
        ntwk.s = np.zeros(shape=(npoints, nports, nports), dtype=complex)

        for m, source_port in enumerate(ports):
            for n, receive_port in enumerate(ports):
                port_key = "S{:d}{:d}".format(receive_port, source_port)
                trace = trace_name[trace_data.index(port_key)]
                self.scpi.set_par_select(channel, trace)
                sdata = self.scpi.query_data(channel, "SDATA")
                ntwk.s[:, m, n] = sdata[::2] + 1j * sdata[1::2]

        name = kwargs.get("name", None)
        if not name:
            port_string = ",".join(map(str, ports))
            name = "{:d}-Port Network ({:})".format(nports, port_string)
        ntwk.name = name
        return ntwk
